import asyncio
from dotenv import load_dotenv
import inspect
import json
import os
import time
from pathlib import Path

from naptha_sdk.client.hub import Hub
from naptha_sdk.client.node import UserClient
from naptha_sdk.configs import setup_module_deployment
from naptha_sdk.inference import InferenceClient
from naptha_sdk.module_manager import AGENT_DIR, add_files_to_package, add_dependencies_to_pyproject, git_add_commit, \
    init_agent_package, publish_ipfs_package, render_agent_code, write_code_to_package
from naptha_sdk.schemas import User
from naptha_sdk.scrape import scrape_init, scrape_func, scrape_func_params
from naptha_sdk.user import get_public_key
from naptha_sdk.utils import get_logger, url_to_node

logger = get_logger(__name__)

load_dotenv(override=True)

class Naptha:
    """The entry point into Naptha."""

    def __init__(self):
        self.public_key = get_public_key(os.getenv("PRIVATE_KEY")) if os.getenv("PRIVATE_KEY") else None
        self.user = User(id=f"user:{self.public_key}")
        self.hub_username = os.getenv("HUB_USERNAME", None)
        self.hub_url = os.getenv("HUB_URL", None)

        node_url = os.getenv("NODE_URL")

        if node_url is None:
            raise ValueError("NODE_URL is not set. Make sure your project has a .env file with a NODE_URL variable.")

        self.node = UserClient(url_to_node(node_url))
        self.inference_client = InferenceClient(url_to_node(node_url))
        self.hub = Hub(self.hub_url, self.public_key)  

    async def __aenter__(self):
        """Async enter method for context manager"""
        await self.hub.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async exit method for context manager"""
        await self.hub.close()

    async def create_agent(self, name):
        async with self.hub:
            _, _, user_id = await self.hub.signin(self.hub_username, os.getenv("HUB_PASSWORD"))
            agent_config = {
                "id": f"agent:{name}",
                "name": name,
                "description": name,
                "author": self.hub.user_id,
                "module_url": "None",
                "module_type": "agent",
                "module_version": "0.1",
                "execution_type": "agent"
            }
            logger.info(f"Registering Agent {agent_config}")
            agent = await self.hub.create_or_update_agent(agent_config)
            if agent:
                logger.info(f"Agent {name} created successfully")
            else:
                logger.error(f"Failed to create agent {name}")

    async def publish_modules(self, decorator = False, register = None, subdeployments = False):
        logger.info(f"Publishing Agent Packages...")
        start_time = time.time()

        if not decorator:
            module_path = Path.cwd()
            deployment_path = module_path / module_path.name / 'configs/deployment.json'
            with open(deployment_path, 'r') as f:
                deployment = json.load(f)
            module = deployment[0]['module']
            modules = [module]
            if subdeployments:
                deployment = await setup_module_deployment(module['module_type'], deployment_path)
                for module_type in ['agent', 'kb', 'tool', 'environment']:
                    subdeployment = module_type + '_deployments'
                    if hasattr(deployment, subdeployment) and getattr(deployment, subdeployment):
                        for submodule in getattr(deployment, subdeployment):
                            modules.append(submodule.module)
        else:
            path = Path.cwd() / AGENT_DIR
            modules = [item.name for item in path.iterdir() if item.is_dir()]
            for module in modules:
                git_add_commit(module)
                module = {
                    "name": module,
                    "description": module,
                    "parameters": "None",
                    "module_type": "agent",
                    "module_url": "None",
                    "module_version": "v0.1",
                    "module_entrypoint": "run.py",
                    "execution_type": "package"
                }
        for module in modules:
            if "module_url" in module and module['module_url'] is not "None":
                module_url = module['module_url']
            # For decorator=False, only the main module should not have a module_url
            else:
                # If register is a string, use it as the URL
                if isinstance(register, str):
                    module_url = register
                    logger.info(f"Using provided URL for {module['module_type']} {module['name']}: {module_url}")
                # Otherwise, publish to IPFS
                else:
                    _, response = await publish_ipfs_package(module['name'], decorator)
                    module_url = f'ipfs://{response["ipfs_hash"]}'
                    logger.info(f"Storing {module['module_type']} {module['name']} on IPFS")
                    logger.info(f"IPFS Hash: {response['ipfs_hash']}. You can download it from http://provider.akash.pro:30584/ipfs/{response['ipfs_hash']}")

            if register:
                # Register module with hub
                async with self.hub:
                    _, _, user_id = await self.hub.signin(self.hub_username, os.getenv("HUB_PASSWORD"))
                    module_config = {
                        "id": f"{module['module_type']}:{module['name']}",
                        "name": module['name'],
                        "description": module['description'],
                        "parameters": module['parameters'],
                        "author": self.hub.user_id,
                        "module_url": module_url,
                        "module_type": module['module_type'],
                        "module_version": module['module_version'],
                        "module_entrypoint": module['module_entrypoint'],
                        "execution_type": module['execution_type'],
                    }
                    logger.info(f"Registering {module['module_type']} {module['name']} on Naptha Hub {module_config}")
                    module = await self.hub.create_or_update_module(module['module_type'], module_config)

        end_time = time.time()
        total_time = end_time - start_time
        logger.info(f"Total time taken to publish {len(modules)} modules: {total_time:.2f} seconds")

    def build(self):
        asyncio.run(self.build_agents())

    async def connect_publish(self):
        await self.hub.connect()
        await self.hub.signin(os.getenv("HUB_USERNAME"), os.getenv("HUB_PASSWORD"))
        await self.publish_agents()
        await self.hub.close()

    def publish(self):
        asyncio.run(self.connect_publish())


def agent(name):
    def decorator(func):
        frame = inspect.currentframe()
        caller_frame = frame.f_back
        instantiation_file = caller_frame.f_code.co_filename
        variables = scrape_init(instantiation_file)
        params = scrape_func_params(func)
        agent_code, obj_name, local_modules, selective_import_modules, standard_import_modules, variable_modules, union_modules = scrape_func(func, variables)
        agent_code = render_agent_code(name, agent_code, obj_name, local_modules, selective_import_modules, standard_import_modules, variable_modules, union_modules, params)
        init_agent_package(name)
        write_code_to_package(name, agent_code)
        add_dependencies_to_pyproject(name, selective_import_modules + standard_import_modules)
        add_files_to_package(name, params, os.getenv("HUB_USERNAME"))

        loop = asyncio.get_event_loop()
        if loop.is_running():
            asyncio.ensure_future(Naptha().create_agent(name))
        else:
            loop.run_until_complete(Naptha().create_agent(name))

        return func
    return decorator

class Agent:
    def __init__(self, 
        name, 
        fn, 
        worker_node_url, 
    ):
        self.name = name
        self.fn = fn
        self.worker_node_url = worker_node_url
        self.repo_id = name


from naptha_sdk.client.node import NodeClient
from naptha_sdk.schemas import EnvironmentDeployment, EnvironmentRunInput
from typing import Any, Dict, List
import logging

logger = logging.getLogger(__name__)

class Environment:
    def __init__(self, environment_deployment: EnvironmentDeployment):
        self.environment_deployment = environment_deployment
        self.environment_node = NodeClient(self.environment_deployment.node)

    async def call_environment_func(self, module_run: EnvironmentRunInput):
        logger.info(f"Running environment on environment node {self.environment_node}")
        environment_run = await self.environment_node.run_module(module_type="environment", run_input=module_run)
        return environment_run



import json
import os
import shutil
import tempfile
import time
import traceback
import uuid
import zipfile
import random
from pathlib import Path
from typing import Dict, Optional, Any, List, Tuple, Union

import grpc
import httpx
import websockets
from google.protobuf import struct_pb2
from google.protobuf.json_format import MessageToDict
from httpx import HTTPStatusError, RemoteProtocolError

from naptha_sdk.client import grpc_server_pb2
from naptha_sdk.client import grpc_server_pb2_grpc
from naptha_sdk.schemas import AgentRun, AgentRunInput, ChatCompletionRequest, EnvironmentRun, EnvironmentRunInput, OrchestratorRun, \
    OrchestratorRunInput, AgentDeployment, EnvironmentDeployment, OrchestratorDeployment, KBDeployment, KBRunInput, KBRun, ToolRunInput, ToolRun, NodeConfig, NodeConfigUser, ModelResponse, ToolDeployment
from naptha_sdk.utils import get_logger, node_to_url

logger = get_logger(__name__)
HTTP_TIMEOUT = 300

class NodeClient:
    def __init__(self, node: NodeConfig):
        self.node = node
        self.server_type = node.server_type
        self.node_url = self.node_to_url(node)
        self.connections = {}

        self.access_token = None
        logger.info(f"Node URL: {self.node_url}")

    def node_to_url(self, node: NodeConfig):
        ports = node.ports
        if len(ports) == 0:
            raise ValueError("No ports found for node")
        if node.server_type == 'ws':
            return f"ws://{node.ip}:{random.choice(ports)}"
        elif node.server_type == 'grpc':
            return f"{node.ip}:{random.choice(ports)}"
        else:
            raise ValueError("Invalid server type. Server type must be either 'ws' or 'grpc'.")

    async def check_user(self, user_input: Dict[str, str]) -> Dict[str, Any]:
        if self.node.server_type == 'ws':
            return await self.check_user_ws(user_input)
        elif self.node.server_type == 'grpc':
            return await self.check_user_grpc(user_input)
        else:
            raise ValueError("Invalid server type. Server type must be either 'ws' or 'grpc'.")

    async def check_user_ws(self, user_input: Dict[str, str]):
        response = await self.send_receive_ws(user_input, "user/check")
        logger.info(f"Check user response: {response}")
        return response

    async def check_user_grpc(self, user_input: Dict[str, str]):
        async with grpc.aio.insecure_channel(self.node_url) as channel:
            stub = grpc_server_pb2_grpc.GrpcServerStub(channel)
            request = grpc_server_pb2.CheckUserRequest(
                user_id=user_input.get('user_id', ''),
                public_key=user_input.get('public_key', '')
            )
            response = await stub.CheckUser(request)
            logger.info(f"Check user response: {response}")
            return MessageToDict(response, preserving_proto_field_name=True)

    async def register_user(self, user_input: Dict[str, str]) -> Dict[str, Any]:
        if self.node.server_type == 'ws':
            return await self.register_user_ws(user_input)
        elif self.node.server_type == 'grpc':
            return await self.register_user_grpc(user_input)
        else:
            raise ValueError("Invalid server type. Server type must be either 'ws' or 'grpc'.")
        
    async def register_user_ws(self, user_input: Dict[str, str]):
        response = await self.send_receive_ws(user_input, "user/register")
        logger.info(f"Register user response: {response}")
        return response

    async def register_user_grpc(self, user_input: Dict[str, str]):
        async with grpc.aio.insecure_channel(self.node_url) as channel:
            stub = grpc_server_pb2_grpc.GrpcServerStub(channel)
            request = grpc_server_pb2.RegisterUserRequest(
                public_key=user_input.get('public_key', '')
            )
            response = await stub.RegisterUser(request)
            return {
                'id': response.id,
                'public_key': response.public_key,
            }

    async def run_module(self, module_type: str, run_input: Union[AgentRunInput, KBRunInput, ToolRunInput, EnvironmentRunInput]):
        if self.node.server_type == 'ws':
            return await self.run_module_ws(module_type, run_input)
        elif self.node.server_type == 'grpc':
            return await self.run_module_grpc(module_type, run_input)
        else:
            raise ValueError("Invalid server type. Server type must be either 'ws' or 'grpc'.")

    async def run_module_ws(self, module_type: str, run_input):
        response = await self.send_receive_ws(run_input, f"{module_type}/run")
        
        output_types = {
            "agent": AgentRun,
            "kb": KBRun,
            "tool": ToolRun,
            "environment": EnvironmentRun
        }
        
        if response['status'] == 'success':
            return output_types[module_type](**response['data'])
        else:
            logger.error(f"Error running {module_type}: {response['message']}")
            logger.error(f"Full traceback: {traceback.format_exc()}")
            raise Exception(response['message'])

    async def run_module_grpc(self, module_type: str, run_input):
        async with grpc.aio.insecure_channel(self.node_url) as channel:
            stub = grpc_server_pb2_grpc.GrpcServerStub(channel)

            # Convert inputs to Struct
            input_struct = struct_pb2.Struct()
            if run_input.inputs:
                if isinstance(run_input.inputs, dict):
                    input_data = run_input.inputs.dict() if hasattr(run_input.inputs, 'dict') else run_input.inputs
                    input_struct.update(input_data)

            # Create node config
            node_config = grpc_server_pb2.NodeConfigUser(
                ip=run_input.deployment.node.ip,
                http_port=run_input.deployment.node.http_port,
                server_type=run_input.deployment.node.server_type
            )

            # Create module
            module = grpc_server_pb2.Module(
                id=run_input.deployment.module.get('id', ''),
                name=run_input.deployment.module.get('name', ''),
                description=run_input.deployment.module.get('description', ''),
                author=run_input.deployment.module.get('author', ''),
                module_url=run_input.deployment.module.get('module_url', ''),
                module_type=module_type,
                module_version=run_input.deployment.module.get('module_version', ''),
                module_entrypoint=run_input.deployment.module.get('module_entrypoint', '')
            )

            # Create config struct
            config_struct = struct_pb2.Struct()
            if run_input.deployment.config:
                if isinstance(run_input.deployment.config, dict):
                    config_struct.update(run_input.deployment.config)
                else:
                    config_struct.update(run_input.deployment.config.dict())

            # Create deployment based on module type
            deployment_classes = {
                "agent": grpc_server_pb2.AgentDeployment,
                "kb": grpc_server_pb2.BaseDeployment,
                "tool": grpc_server_pb2.ToolDeployment,
                "environment": grpc_server_pb2.BaseDeployment
            }
            
            DeploymentClass = deployment_classes[module_type]
            deployment = DeploymentClass(
                node_input=node_config,
                name=run_input.deployment.name,
                module=module,
                config=config_struct,
                initialized=False
            )

            # Create request with appropriate deployment field
            request_args = {
                "module_type": module_type,
                "consumer_id": run_input.consumer_id,
                "inputs": input_struct,
                f"{module_type}_deployment": deployment
            }
            
            request = grpc_server_pb2.ModuleRunRequest(**request_args)

            final_response = None
            async for response in stub.RunModule(request):
                final_response = response
                logger.info(f"Got response: {response}")

            output_types = {
                "agent": AgentRun,
                "kb": KBRun,
                "tool": ToolRun,
                "environment": EnvironmentRun
            }

            return output_types[module_type](
                consumer_id=run_input.consumer_id,
                inputs=run_input.inputs,
                deployment=run_input.deployment,
                orchestrator_runs=[],
                status=final_response.status,
                error=final_response.error,
                id=final_response.id,
                results=list(final_response.results),
                error_message=final_response.error_message,
                created_time=final_response.created_time,
                start_processing_time=final_response.start_processing_time,
                completed_time=final_response.completed_time,
                duration=final_response.duration
            )
    
    async def connect_ws(self, action: str):
        client_id = str(uuid.uuid4())
        full_url = f"{self.node_url}/ws/{action}/{client_id}"
        logger.info(f"Connecting to WebSocket: {full_url}")
        ws = await websockets.connect(full_url)
        self.connections[client_id] = ws
        self.current_client_id = client_id
        return client_id

    async def disconnect_ws(self, client_id: str):
        if client_id in self.connections:
            await self.connections[client_id].close()
            del self.connections[client_id]
        if self.current_client_id == client_id:
            self.current_client_id = None

    async def send_receive_ws(self, data, action: str):
        client_id = await self.connect_ws(action)
        
        try:
            if isinstance(data, AgentRunInput) or isinstance(data, OrchestratorRunInput):
                message = data.model_dump()
            else:
                message = data
            await self.connections[client_id].send(json.dumps(message))
            
            response = await self.connections[client_id].recv()
            return json.loads(response)
        finally:
            await self.disconnect_ws(client_id)

class UserClient:
    def __init__(self, node: NodeConfigUser):
        self.node = node
        self.node_url = node_to_url(node)
        self.connections = {}
        
        self.access_token = None
        logger.info(f"Node URL: {self.node_url}")

    async def create(self, module_type: str,
                     module_request: Union[AgentDeployment, EnvironmentDeployment, KBDeployment, OrchestratorDeployment, ToolDeployment]):
        """Generic method to create either an agent, orchestrator, environment, tool, kb or memory.

        Args:
            module_type: Either agent, orchestrator, environment, tool, kb or memory
            module_request: Either AgentDeployment, EnvironmentDeployment, OrchestratorDeployment, ToolDeployment, KBDeployment or MemoryDeployment
        """

        print(f"Creating {module_type}...")

        endpoint = f"{self.node_url}/{module_type}/create"
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.access_token}',
                }
                response = await client.post(
                    endpoint,
                    json=module_request.model_dump(),
                    headers=headers
                )
                response.raise_for_status()

                # Convert response to appropriate return type
                return response.json()
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise
        except RemoteProtocolError as e:
            error_msg = f"Run {module_type} failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.error(error_msg)
            raise
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            raise

    async def _run_and_poll(self, run_input: Union[AgentRunInput, EnvironmentRunInput, OrchestratorRunInput, KBRunInput, ToolRunInput, Dict], module_type: str) -> Union[AgentRun, EnvironmentRun, OrchestratorRun, KBRun, ToolRun, Dict]:
        """Generic method to run and poll either an agent, orchestrator, environment, tool or KB.
        
        Args:
            run_input: Either AgentRunInput, OrchestratorRunInput, EnvironmentRunInput, KBRunInput, ToolRunInput or Dict
            module_type: Either 'agent', 'orchestrator', 'environment', 'tool' or 'kb'
        """
        print(f"Run input: {run_input}")
        print(f"Module type: {module_type}")
        # Start the run
        run = await getattr(self, f'run_{module_type}')(run_input)
        print(f"{module_type.title()} run started: {run}")

        current_results_len = 0
        while True:
            run = await getattr(self, f'check_{module_type}_run')(run)

            output = f"{run.status} {getattr(run, f'deployment').module['module_type']} {getattr(run, f'deployment').module['name']}"
            print(output)

            results = run.results
            status = run.status

            if len(results) > current_results_len:
                print("Output: ", results[-1])
                current_results_len += 1

            if status in ['completed', 'error']:
                break

            time.sleep(3)

        if status == 'completed':
            print(results)
        else:
            error_msg = run.error_message
            print(error_msg)
        return run

    async def run_agent_and_poll(self, agent_run_input: AgentRunInput) -> AgentRun:
        """Run an agent and poll for results until completion."""
        return await self._run_and_poll(agent_run_input, 'agent')

    async def run_tool_and_poll(self, tool_run_input: ToolRunInput) -> ToolRun:
        """Run a tool and poll for results until completion."""

        return await self._run_and_poll(tool_run_input, 'tool')

    async def run_orchestrator_and_poll(self, orchestrator_run_input: OrchestratorRunInput) -> OrchestratorRun:
        """Run an orchestrator and poll for results until completion."""
        return await self._run_and_poll(orchestrator_run_input, 'orchestrator')

    async def run_environment_and_poll(self, environment_input: EnvironmentRunInput) -> EnvironmentRun:
        """Run an environment and poll for results until completion."""
        return await self._run_and_poll(environment_input, 'environment')
    
    async def run_kb_and_poll(self, kb_input: KBDeployment) -> KBDeployment:
        """Run a knowledge base and poll for results until completion."""
        return await self._run_and_poll(kb_input, 'kb')

    async def check_user(self, user_input: Dict[str, Any]) -> Dict[str, Any]:
        """
        Check if a user exists on a node
        """
        endpoint = self.node_url + "/user/check"
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json', 
                }
                response = await client.post(
                    endpoint, 
                    json=user_input,
                    headers=headers
                )
                response.raise_for_status()
            return json.loads(response.text)
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except RemoteProtocolError as e:
            error_msg = f"Check user failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.info(error_msg)
            raise 
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")
            raise

    async def register_user(self, user_input: Dict[str, Any]) -> Dict[str, Any]:
        """
        Register a user on a node
        """
        endpoint = self.node_url + "/user/register"
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json', 
                }
                response = await client.post(
                    endpoint, 
                    json=user_input,
                    headers=headers
                )
                response.raise_for_status()
            return json.loads(response.text)
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except RemoteProtocolError as e:
            error_msg = f"Register user failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.error(error_msg)
            raise 
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")
            raise

    async def _run_module(self, run_input: Union[AgentRunInput, OrchestratorRunInput, EnvironmentRunInput, ToolRunInput], module_type: str) -> Union[AgentRun, OrchestratorRun, EnvironmentRun, ToolRun]:
        """
        Generic method to run either an agent, orchestrator, environment, or tool on a node
        
        Args:
            run_input: Either AgentRunInput, OrchestratorRunInput, EnvironmentRunInput, or ToolRunInput
            module_type: Either 'agent', 'orchestrator', 'environment', or 'tool'
        """
        print(f"Running {module_type}...")
        print(f"Run input: {run_input}")
        print(f"Node URL: {self.node_url}")

        endpoint = f"{self.node_url}/{module_type}/run"
        
        # Convert dict to appropriate input type if needed
        input_class = {
            'agent': AgentRunInput,
            'orchestrator': OrchestratorRunInput,
            'environment': EnvironmentRunInput,
            'kb': KBRunInput,
            'tool': ToolRunInput
        }[module_type]
        
        if isinstance(run_input, dict):
            run_input = input_class(**run_input)

        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.access_token}',
                }
                response = await client.post(
                    endpoint,
                    json=run_input.model_dict(),
                    headers=headers
                )

                # Try to get error details even for error responses
                if response.status_code >= 400:
                    error_detail = response.json() if response.text else str(response)
                    logger.error(f"Server error response: {error_detail}")
                    raise Exception(f"Server returned error response: {error_detail}")
                    
                response.raise_for_status()
                
                # Convert response to appropriate return type
                return_class = {
                    'agent': AgentRun,
                    'orchestrator': OrchestratorRun,
                    'environment': EnvironmentRun,
                    'kb': KBRun,
                    'tool': ToolRun
                }[module_type]
                return return_class(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise
        except RemoteProtocolError as e:
            error_msg = f"Run {module_type} failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.error(error_msg)
            raise
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            raise

    async def run_inference(self, inference_input: Union[ChatCompletionRequest, Dict]) -> ModelResponse:
        """
        Run inference on a node
        
        Args:
            inference_input: The inference input to run inference on
        """
        if isinstance(inference_input, dict):
            inference_input = ChatCompletionRequest(**inference_input)

        endpoint = f"{self.node_url}/inference/chat"

        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.access_token}',
                }
                response = await client.post(
                    endpoint,
                    json=inference_input.model_dump(),
                    headers=headers
                )
                print("Response: ", response.text)
                response.raise_for_status()
                return ModelResponse(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise
        except RemoteProtocolError as e:
            error_msg = f"Inference failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.error(error_msg)
            raise
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            raise

    async def run_agent(self, agent_run_input: AgentRunInput) -> AgentRun:
        """Run an agent on a node"""
        return await self._run_module(agent_run_input, 'agent')

    async def run_tool(self, tool_run_input: ToolRunInput) -> ToolRun:
        """Run a tool on a node"""
        return await self._run_module(tool_run_input, 'tool')

    async def run_orchestrator(self, orchestrator_run_input: OrchestratorRunInput) -> OrchestratorRun:
        """Run an orchestrator on a node"""
        return await self._run_module(orchestrator_run_input, 'orchestrator')
    
    async def run_environment(self, environment_run_input: EnvironmentRunInput) -> EnvironmentRun:
        """Run an environment on a node"""
        return await self._run_module(environment_run_input, 'environment')

    async def run_kb(self, kb_run_input: KBRunInput) -> KBRun:
        """Run a knowledge base on a node"""
        return await self._run_module(kb_run_input, 'kb')

    async def check_run(
        self, 
        module_run: Union[AgentRun, OrchestratorRun, EnvironmentRun, KBRun, ToolRun], 
        module_type: str
    ) -> Union[AgentRun, OrchestratorRun, EnvironmentRun, KBRun, ToolRun]:
        """Generic method to check the status of a module run.
        
        Args:
            module_run: Either AgentRun, OrchestratorRun, EnvironmentRun, ToolRun or KBRun object
            module_type: Either 'agent', 'orchestrator', 'environment', 'tool' or 'kb'
        """
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                response = await client.post(
                    f"{self.node_url}/{module_type}/check", 
                    json=module_run.model_dump()
                )
                response.raise_for_status()
            
            return_class = {
                'agent': AgentRun,
                'orchestrator': OrchestratorRun,
                'environment': EnvironmentRun,
                'kb': KBRun,
                'tool': ToolRun
            }[module_type]
            return return_class(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")

    # Update existing methods to use the new generic one
    async def check_agent_run(self, agent_run: AgentRun) -> AgentRun:
        return await self.check_run(agent_run, 'agent')

    async def check_tool_run(self, tool_run: ToolRun) -> ToolRun:
        return await self.check_run(tool_run, 'tool')

    async def check_orchestrator_run(self, orchestrator_run: OrchestratorRun) -> OrchestratorRun:
        return await self.check_run(orchestrator_run, 'orchestrator')

    async def check_environment_run(self, environment_run: EnvironmentRun) -> EnvironmentRun:
        return await self.check_run(environment_run, 'environment')

    async def check_kb_run(self, kb_run: KBRun) -> KBRun:
        return await self.check_run(kb_run, 'kb')

    async def create_agent_run(self, agent_run_input: AgentRunInput) -> AgentRun:
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                response = await client.post(
                    f"{self.node_url}/monitor/create_agent_run", json=agent_run_input.model_dump()
                )
                response.raise_for_status()
            return AgentRun(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")
            logger.info(f"Full traceback: {traceback.format_exc()}")

    async def update_agent_run(self, agent_run: AgentRun):
        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                response = await client.post(
                    f"{self.node_url}/monitor/update_agent_run", json=agent_run.model_dump()
                )
                response.raise_for_status()
            return AgentRun(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            error_details = traceback.format_exc()
            print(f"Full traceback: {error_details}")

    async def read_storage(self, agent_run_id: str, output_dir: str, ipfs: bool = False) -> str:
        print("Reading from storage...")
        try:
            endpoint = f"{self.node_url}/{'storage/read_ipfs' if ipfs else 'storage/read'}/{agent_run_id}"

            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                response = await client.get(endpoint)
                response.raise_for_status()
                storage = response.content  
                print("Retrieved storage.")
            
                # Temporary file handling
                temp_file_name = None
                with tempfile.NamedTemporaryFile(delete=False, mode='wb') as tmp_file:
                    tmp_file.write(storage)  # storage is a bytes-like object
                    temp_file_name = tmp_file.name
        
                # Ensure output directory exists
                output_path = Path(output_dir)
                output_path.mkdir(parents=True, exist_ok=True)
        
                # Check if the file is a zip file and extract if true
                if zipfile.is_zipfile(temp_file_name):
                    with zipfile.ZipFile(temp_file_name, 'r') as zip_ref:
                        zip_ref.extractall(output_path)
                    print(f"Extracted storage to {output_dir}.")
                else:
                    shutil.copy(temp_file_name, output_path)
                    print(f"Copied storage to {output_dir}.")

                # Cleanup temporary file
                Path(temp_file_name).unlink(missing_ok=True)
        
                return output_dir         
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")
            logger.info(f"Full traceback: {traceback.format_exc()}")

    async def write_storage(self, storage_input: str, ipfs: bool = False, publish_to_ipns: bool = False, update_ipns_name: str = None) -> Dict[str, Any]:
        """Write storage to the node."""
        print("Writing storage")
        try:
            file = prepare_files(storage_input)
            endpoint = f"{self.node_url}/storage/write_ipfs" if ipfs else f"{self.node_url}/storage/write"
            
            if update_ipns_name:
                publish_to_ipns = True

            data = {
                "publish_to_ipns": publish_to_ipns,
                "update_ipns_name": update_ipns_name
            }
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                response = await client.post(
                    endpoint, 
                    files=file,
                    data=data,
                    timeout=600
                )
                response.raise_for_status()
                return response.json()
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise  
        except Exception as e:
            logger.info(f"An unexpected error occurred: {e}")
            logger.info(f"Full traceback: {traceback.format_exc()}")
            return {}

    async def create_table(self, table_name: str, schema: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.post(
                f"{self.node_url}/local-db/create-table",
                json={"table_name": table_name, "schema": schema}
            )
            response.raise_for_status()
            return response.json()

    async def add_row(self, table_name: str, data: Dict[str, Any], schema: Optional[Dict[str, Dict[str, Any]]] = None) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.post(
                f"{self.node_url}/local-db/add-row",
                json={"table_name": table_name, "data": data, "schema": schema}
            )
            response.raise_for_status()
            return response.json()

    async def update_row(self, table_name: str, data: Dict[str, Any], condition: Dict[str, Any], schema: Optional[Dict[str, Dict[str, Any]]] = None) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.post(
                f"{self.node_url}/local-db/update-row",
                json={
                    "table_name": table_name,
                    "data": data,
                    "condition": condition,
                    "schema": schema
                }
            )
            response.raise_for_status()
            return response.json()

    async def delete_row(self, table_name: str, condition: Dict[str, Any]) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.post(
                f"{self.node_url}/local-db/delete-row",
                json={"table_name": table_name, "condition": condition}
            )
            response.raise_for_status()
            return response.json()

    async def list_tables(self) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.get(f"{self.node_url}/local-db/tables")
            response.raise_for_status()
            return response.json()

    async def get_table_schema(self, table_name: str) -> Dict[str, Any]:
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.get(f"{self.node_url}/local-db/table/{table_name}")
            response.raise_for_status()
            return response.json()

    async def query_table(self, table_name: str, columns: Optional[str] = None, condition: Optional[Union[str, Dict]] = None, order_by: Optional[str] = None, limit: Optional[int] = None) -> Dict[str, Any]:
        params = {"table_name": table_name}
        if columns:
            params["columns"] = columns
        if condition:
            params["condition"] = json.dumps(condition) if isinstance(condition, dict) else condition
        if order_by:
            params["order_by"] = order_by
        if limit:
            params["limit"] = limit

        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.get(
                f"{self.node_url}/local-db/table/{table_name}/rows",
                params=params
            )
            response.raise_for_status()
            return response.json()

    async def vector_search(
        self,
        table_name: str,
        vector_column: str,
        query_vector: List[float],
        columns: Optional[List[str]] = None,
        top_k: int = 5,
        include_similarity: bool = True
    ) -> Dict[str, Any]:
        """
        Perform a pgvector-based similarity search on a table's vector column.
        """
        payload = {
            "table_name": table_name,
            "vector_column": vector_column,
            "query_vector": query_vector,
            "columns": columns or ["text"],
            "top_k": top_k,
            "include_similarity": include_similarity,
        }
        async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
            response = await client.post(
                f"{self.node_url}/local-db/vector_search",
                json=payload
            )
            response.raise_for_status()
            return response.json()


def zip_directory(file_path, zip_path):
    """Utility function to zip the content of a directory while preserving the folder structure."""
    with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(file_path):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, start=os.path.abspath(file_path).split(os.sep)[0])
                zipf.write(file_path, arcname)

def prepare_files(file_path: str) -> List[Tuple[str, str]]:
    """Prepare files for upload."""
    if os.path.isdir(file_path):
        with tempfile.NamedTemporaryFile(delete=False, suffix='.zip') as tmpfile:
            zip_directory(file_path, tmpfile.name)
            tmpfile.close()  
            file = {'file': open(tmpfile.name, 'rb')}
    else:
        file = {'file': open(file_path, 'rb')}
    
    return file




import httpx
import json
from pydantic import BaseModel
from typing import Union, Dict, Any, Optional, List, BinaryIO
from naptha_sdk.schemas import NodeConfigUser
from naptha_sdk.storage.schemas import (
    StorageLocation,
    StorageType,
    StorageObject,
    BaseStorageRequest,
    CreateStorageRequest,
    ReadStorageRequest,
    UpdateStorageRequest,
    DeleteStorageRequest,
    ListStorageRequest,
    SearchStorageRequest
)
from naptha_sdk.utils import get_logger, node_to_url

HTTP_TIMEOUT = 300

logger = get_logger(__name__)

class StorageProvider:
    def __init__(self, node: NodeConfigUser):
        self.node = node
        self.node_url = node_to_url(node)
        self.client = httpx.AsyncClient(timeout=HTTP_TIMEOUT)
        logger.info(f"Storage Provider URL: {self.node_url}")


    async def _make_request(
        self,
        request: BaseStorageRequest,
        files: Optional[Dict] = None
    ) -> Any:
        """Make HTTP request to storage endpoint"""
        endpoint = f"{self.node_url}/storage/{request.storage_type.value}/{request.request_type.value}/{request.path}"
        print(f"Request: {request}")
        try:
            response = None
            match request:
                case CreateStorageRequest():
                    form_data = {}
                    if not request.data:
                        request.data = {}
                    if not request.options:
                        request.options = {}
                    form_data['data'] = json.dumps({**request.data, **request.options})
                    print(f"Form data: {form_data}")
                    files = files if files is not None else {}
                    response = await self.client.post(endpoint, data=form_data, files=files)

                case ReadStorageRequest():
                    if request.storage_type in [StorageType.FILESYSTEM, StorageType.IPFS]:
                        response = await self.client.get(endpoint)
                        response.raise_for_status()
                        # Check content type to determine response handling
                        content_type = response.headers.get('content-type', '')
                        if 'json' in content_type:
                            return response.json()
                        return response.content
                    else:
                        params = {"options": json.dumps(request.options)} if request.options else None
                        response = await self.client.get(endpoint, params=params)
                                        
                case UpdateStorageRequest():
                    # Extract condition from options if present
                    condition = request.options.get("condition") if request.options else None
                    form_data = {
                        'data': json.dumps(request.data)
                    }
                    params = {
                        'condition': json.dumps(condition) if condition else None
                    }
                    response = await self.client.put(endpoint, data=form_data, params=params)
                    
                case ListStorageRequest():
                    params = {"options": json.dumps(request.options)} if request.options else None
                    response = await self.client.get(endpoint, params=params)
                    
                case DeleteStorageRequest():
                    params = {}
                    if request.storage_type == StorageType.DATABASE:
                        if request.condition:
                            params["condition"] = json.dumps(request.condition)
                    elif request.storage_type == StorageType.FILESYSTEM:
                        if request.options:
                            params["options"] = json.dumps(request.options)
                    response = await self.client.delete(endpoint, params=params)
                    
                case SearchStorageRequest():
                    search_data = {
                        "query": request.query,
                        "query_type": request.query_type,
                        "limit": request.limit
                    }
                    response = await self.client.post(endpoint, json=search_data)
            
            if response:
                response.raise_for_status()
                content_type = response.headers.get('content-type', '')
                if 'json' in content_type:
                    return response.json()
                return response.content
                
        except httpx.HTTPStatusError as e:
            logger.error(f"HTTP error occurred: {e.response.text}")
            raise StorageError(f"HTTP error occurred: {str(e)}", status_code=e.response.status_code)
        except Exception as e:
            logger.error(f"Storage operation failed: {str(e)}")
            raise StorageError(f"Storage operation failed: {str(e)}")

    async def execute(self, request: BaseStorageRequest) -> Union[StorageObject, List[StorageObject], bool]:
        """Execute storage request and return appropriate response"""
        files = None
        if isinstance(request, CreateStorageRequest) and request.file:
            files = {"file": request.file}
            
        result = await self._make_request(request, files=files)

        match request:
            case DeleteStorageRequest():
                return True
                
            case ListStorageRequest():
                # Handle list response which might be a simple array
                if isinstance(result, list):
                    return [
                        StorageObject(
                            location=StorageLocation(
                                storage_type=request.storage_type,
                                path=request.path
                            ),
                            data=item
                        )
                        for item in result
                    ]
                return StorageObject(
                    location=StorageLocation(storage_type=request.storage_type, path=request.path),
                    data=result
                )
                
            case SearchStorageRequest():
                # Handle search response which returns objects with path
                return [
                    StorageObject(
                        location=StorageLocation(
                            storage_type=request.storage_type,
                            path=obj.get("path", "")
                        ),
                        data=obj.get("data"),
                        metadata=obj.get("metadata")
                    )
                    for obj in result
                ]
                
            case _:
                return StorageObject(
                    location=StorageLocation(
                        storage_type=request.storage_type,
                        path=request.path
                    ),
                    data=result
                )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.aclose()

class StorageError(Exception):
    """Custom exception for storage operations"""
    def __init__(self, message: str, status_code: Optional[int] = None):
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)


import logging
import os
import yaml
from naptha_sdk.schemas import NodeConfigUser

def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

def load_yaml(cfg_path):
    with open(cfg_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.FullLoader)
    return cfg

def add_credentials_to_env(username, password, private_key_path):
    env_file_path = os.path.join(os.getcwd(), '.env')
    updated_lines = []
    hub_user_found = False
    hub_pass_found = False
    private_key_found = False

    # Read the existing .env file
    with open(env_file_path, 'r') as env_file:
        for line in env_file:
            if line.startswith('HUB_USERNAME='):
                updated_lines.append(f"HUB_USERNAME={username}\n")
                hub_user_found = True
            elif line.startswith('HUB_PASSWORD='):
                updated_lines.append(f"HUB_PASSWORD={password}\n")
                hub_pass_found = True
            elif line.startswith('PRIVATE_KEY='):
                updated_lines.append(f"PRIVATE_KEY={private_key_path}\n")
                private_key_found = True
            else:
                updated_lines.append(line)

    # Append new credentials if not found
    if not hub_user_found:
        updated_lines.append(f"HUB_USERNAME={username}\n")
    if not hub_pass_found:
        updated_lines.append(f"HUB_PASSWORD={password}\n")
    if not private_key_found:
        updated_lines.append(f"PRIVATE_KEY={private_key_path}\n")

    # Write the updated content back to the .env file
    with open(env_file_path, 'w') as env_file:
        env_file.writelines(updated_lines)

    print("Your credentials have been updated in the .env file. You can now use these credentials to authenticate in future sessions.")

def write_private_key_to_file(private_key, username):
    private_key_file_path = os.path.join(os.getcwd(), f'{username}.pem')
    with open(private_key_file_path, 'w') as file:
        file.write(private_key)
    
    update_private_key_in_env(private_key_file_path)    

def update_private_key_in_env(private_key_path):
    env_file_path = os.path.join(os.getcwd(), '.env')
    updated_lines = []
    private_key_found = False

    with open(env_file_path, 'r') as env_file:
        for line in env_file:
            if line.startswith('PRIVATE_KEY='):
                updated_lines.append(f"PRIVATE_KEY={private_key_path}\n")
                private_key_found = True
            else:
                updated_lines.append(line)

    if not private_key_found:
        updated_lines.append(f"PRIVATE_KEY={private_key_path}\n")

    with open(env_file_path, 'w') as env_file:
        env_file.writelines(updated_lines)

    print("Your private key have been updated in the .env file")

class AsyncMixin:
    def __init__(self, *args, **kwargs):
        """
        Standard constructor used for arguments pass
        Do not override. Use __ainit__ instead
        """
        self.__storedargs = args, kwargs
        self.async_initialized = False

    async def __ainit__(self, *args, **kwargs):
        """Async constructor, you should implement this"""

    async def __initobj(self):
        """Crutch used for __await__ after spawning"""
        assert not self.async_initialized
        self.async_initialized = True
        # pass the parameters to __ainit__ that passed to __init__
        await self.__ainit__(*self.__storedargs[0], **self.__storedargs[1])
        return self

    def __await__(self):
        return self.__initobj().__await__()
    
def node_to_url(node_schema: NodeConfigUser):
    return f"http://{node_schema.ip}:{node_schema.http_port}"
    
def url_to_node(url: str):
    protocol = url.split('://')[0]
    host = url.split('://')[1].split(':')[0] 
    http_port = int(url.split(':')[-1])
    return NodeConfigUser(ip=host, http_port=http_port, server_type=protocol)


from dotenv import load_dotenv
from naptha_sdk.client.node import NodeClient
from naptha_sdk.schemas import AgentRunInput
from naptha_sdk.utils import get_logger

logger = get_logger(__name__)
load_dotenv(override=True)

class Agent:
    def __init__(self, 
        deployment, 
        *args,
        **kwargs
    ):
        self.deployment = deployment
        self.agent_node = NodeClient(self.deployment.node)

    async def call_agent_func(self, module_run_input: AgentRunInput, *args, **kwargs):
        logger.info(f"Running agent on worker node {self.agent_node.node_url}")
        
        agent_run = await self.agent_node.run_module(module_type="agent", run_input=module_run_input.model_dict())
        return agent_run



import json
from typing import Dict, Union
import httpx
from httpx import HTTPStatusError, RemoteProtocolError
from naptha_sdk.schemas import ChatCompletionRequest, NodeConfigUser, ModelResponse
from naptha_sdk.utils import get_logger, node_to_url

logger = get_logger(__name__)
HTTP_TIMEOUT = 300


class InferenceClient:
    def __init__(self, node: NodeConfigUser):
        self.node = node
        self.node_url = node_to_url(node)
        
        self.access_token = None
        logger.info(f"Node URL: {self.node_url}")

    async def run_inference(self, inference_input: Union[ChatCompletionRequest, Dict]) -> Dict:
        """
        Run inference on a node
        
        Args:
            inference_input: The inference input to run inference on
        """
        if isinstance(inference_input, dict):
            inference_input = ChatCompletionRequest(**inference_input)

        endpoint = f"{self.node_url}/inference/chat"

        try:
            async with httpx.AsyncClient(timeout=HTTP_TIMEOUT) as client:
                headers = {
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.access_token}',
                }
                response = await client.post(
                    endpoint,
                    json=inference_input.model_dump(),
                    headers=headers
                )
                print("Response: ", response.text)
                response.raise_for_status()
                return ModelResponse(**json.loads(response.text))
        except HTTPStatusError as e:
            logger.info(f"HTTP error occurred: {e}")
            raise
        except RemoteProtocolError as e:
            error_msg = f"Inference failed to connect to the server at {self.node_url}. Please check if the server URL is correct and the server is running. Error details: {str(e)}"
            logger.error(error_msg)
            raise
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            raise


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: grpc_server.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'grpc_server.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11grpc_server.proto\x12\x05\x61gent\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgoogle/protobuf/empty.proto\".\n\x0fGeneralResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"\xe9\x01\n\tLLMConfig\x12\x18\n\x0b\x63onfig_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06\x63lient\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x12\n\x05model\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x17\n\nmax_tokens\x18\x04 \x01(\x05H\x03\x88\x01\x01\x12\x18\n\x0btemperature\x18\x05 \x01(\x02H\x04\x88\x01\x01\x12\x15\n\x08\x61pi_base\x18\x06 \x01(\tH\x05\x88\x01\x01\x42\x0e\n\x0c_config_nameB\t\n\x07_clientB\x08\n\x06_modelB\r\n\x0b_max_tokensB\x0e\n\x0c_temperatureB\x0b\n\t_api_base\"@\n\nNodeServer\x12\x13\n\x0bserver_type\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x0f\n\x07node_id\x18\x03 \x01(\t\"m\n\x0fNodeConfigInput\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x16\n\thttp_port\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x18\n\x0bserver_type\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\x0c\n\n_http_portB\x0e\n\x0c_server_type\"\xc7\x03\n\nNodeConfig\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t\x12\x12\n\npublic_key\x18\x03 \x01(\t\x12\n\n\x02ip\x18\x04 \x01(\t\x12\x13\n\x0bserver_type\x18\x05 \x01(\t\x12\x11\n\thttp_port\x18\x06 \x01(\x05\x12\x13\n\x0bnum_servers\x18\x07 \x01(\x05\x12\"\n\x07servers\x18\x08 \x03(\x0b\x32\x11.agent.NodeServer\x12\x15\n\rollama_models\x18\t \x03(\t\x12\x13\n\x0b\x64ocker_jobs\x18\n \x01(\x08\x12\r\n\x05ports\x18\x0b \x03(\x05\x12\x19\n\x0crouting_type\x18\x0c \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0brouting_url\x18\r \x01(\tH\x01\x88\x01\x01\x12\x15\n\x08num_gpus\x18\x0e \x01(\x05H\x02\x88\x01\x01\x12\x11\n\x04\x61rch\x18\x0f \x01(\tH\x03\x88\x01\x01\x12\x0f\n\x02os\x18\x10 \x01(\tH\x04\x88\x01\x01\x12\x10\n\x03ram\x18\x11 \x01(\x03H\x05\x88\x01\x01\x12\x11\n\x04vram\x18\x12 \x01(\x03H\x06\x88\x01\x01\x42\x0f\n\r_routing_typeB\x0e\n\x0c_routing_urlB\x0b\n\t_num_gpusB\x07\n\x05_archB\x05\n\x03_osB\x06\n\x04_ramB\x07\n\x05_vram\"\x9b\x02\n\x06Module\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\x12\x12\n\nmodule_url\x18\x05 \x01(\t\x12\x18\n\x0bmodule_type\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x1b\n\x0emodule_version\x18\x07 \x01(\tH\x01\x88\x01\x01\x12\x1e\n\x11module_entrypoint\x18\x08 \x01(\tH\x02\x88\x01\x01\x12\x1b\n\x0e\x65xecution_type\x18\t \x01(\tH\x03\x88\x01\x01\x42\x0e\n\x0c_module_typeB\x11\n\x0f_module_versionB\x14\n\x12_module_entrypointB\x11\n\x0f_execution_type\"\xd0\x02\n\x14\x44\x61taGenerationConfig\x12\x19\n\x0csave_outputs\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\"\n\x15save_outputs_location\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1e\n\x11save_outputs_path\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x18\n\x0bsave_inputs\x18\x04 \x01(\x08H\x03\x88\x01\x01\x12!\n\x14save_inputs_location\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x1d\n\x10\x64\x65\x66\x61ult_filename\x18\x06 \x01(\tH\x05\x88\x01\x01\x42\x0f\n\r_save_outputsB\x18\n\x16_save_outputs_locationB\x14\n\x12_save_outputs_pathB\x0e\n\x0c_save_inputsB\x17\n\x15_save_inputs_locationB\x13\n\x11_default_filename\"\xdb\x01\n\x0e\x42\x61seDeployment\x12(\n\x0bnode_config\x18\x01 \x01(\x0b\x32\x11.agent.NodeConfigH\x00\x12,\n\nnode_input\x18\x02 \x01(\x0b\x32\x16.agent.NodeConfigInputH\x00\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1d\n\x06module\x18\x04 \x01(\x0b\x32\r.agent.Module\x12\'\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x13\n\x0binitialized\x18\x06 \x01(\x08\x42\x06\n\x04node\"\xb8\x02\n\x0eToolDeployment\x12(\n\x0bnode_config\x18\x01 \x01(\x0b\x32\x11.agent.NodeConfigH\x00\x12,\n\nnode_input\x18\x02 \x01(\x0b\x32\x16.agent.NodeConfigInputH\x00\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1d\n\x06module\x18\x04 \x01(\x0b\x32\r.agent.Module\x12\'\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12@\n\x16\x64\x61ta_generation_config\x18\x06 \x01(\x0b\x32\x1b.agent.DataGenerationConfigH\x01\x88\x01\x01\x12\x13\n\x0binitialized\x18\x07 \x01(\x08\x42\x06\n\x04nodeB\x19\n\x17_data_generation_config\"\xd1\x03\n\x0f\x41gentDeployment\x12(\n\x0bnode_config\x18\x01 \x01(\x0b\x32\x11.agent.NodeConfigH\x00\x12,\n\nnode_input\x18\x02 \x01(\x0b\x32\x16.agent.NodeConfigInputH\x00\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1d\n\x06module\x18\x04 \x01(\x0b\x32\r.agent.Module\x12\'\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12@\n\x16\x64\x61ta_generation_config\x18\x06 \x01(\x0b\x32\x1b.agent.DataGenerationConfigH\x01\x88\x01\x01\x12/\n\x10tool_deployments\x18\x07 \x03(\x0b\x32\x15.agent.ToolDeployment\x12-\n\x0ekb_deployments\x18\x08 \x03(\x0b\x32\x15.agent.BaseDeployment\x12\x36\n\x17\x65nvironment_deployments\x18\t \x03(\x0b\x32\x15.agent.BaseDeployment\x12\x13\n\x0binitialized\x18\n \x01(\x08\x42\x06\n\x04nodeB\x19\n\x17_data_generation_config\"\xa3\x03\n\x10ModuleRunRequest\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x13\n\x0b\x63onsumer_id\x18\x02 \x01(\t\x12\'\n\x06inputs\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x32\n\x10\x61gent_deployment\x18\x04 \x01(\x0b\x32\x16.agent.AgentDeploymentH\x00\x12\x30\n\x0ftool_deployment\x18\x05 \x01(\x0b\x32\x15.agent.ToolDeploymentH\x00\x12\x32\n\x11memory_deployment\x18\x06 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12.\n\rkb_deployment\x18\x07 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12\x37\n\x16\x65nvironment_deployment\x18\x08 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12+\n\x11orchestrator_runs\x18\t \x03(\x0b\x32\x10.agent.ModuleRunB\x0c\n\ndeployment\"\x90\x06\n\tModuleRun\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x13\n\x0b\x63onsumer_id\x18\x02 \x01(\t\x12\'\n\x06inputs\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x32\n\x10\x61gent_deployment\x18\x04 \x01(\x0b\x32\x16.agent.AgentDeploymentH\x00\x12\x30\n\x0ftool_deployment\x18\x05 \x01(\x0b\x32\x15.agent.ToolDeploymentH\x00\x12\x32\n\x11memory_deployment\x18\x06 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12.\n\rkb_deployment\x18\x07 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12\x37\n\x16\x65nvironment_deployment\x18\x08 \x01(\x0b\x32\x15.agent.BaseDeploymentH\x00\x12+\n\x11orchestrator_runs\x18\t \x03(\x0b\x32\x10.agent.ModuleRun\x12\x0e\n\x06status\x18\n \x01(\t\x12\r\n\x05\x65rror\x18\x0b \x01(\x08\x12\x0f\n\x02id\x18\x0c \x01(\tH\x01\x88\x01\x01\x12\x0f\n\x07results\x18\r \x03(\t\x12\x1a\n\rerror_message\x18\x0e \x01(\tH\x02\x88\x01\x01\x12\x19\n\x0c\x63reated_time\x18\x0f \x01(\tH\x03\x88\x01\x01\x12\"\n\x15start_processing_time\x18\x10 \x01(\tH\x04\x88\x01\x01\x12\x1b\n\x0e\x63ompleted_time\x18\x11 \x01(\tH\x05\x88\x01\x01\x12\x15\n\x08\x64uration\x18\x12 \x01(\x01H\x06\x88\x01\x01\x12#\n\x16input_schema_ipfs_hash\x18\x13 \x01(\tH\x07\x88\x01\x01\x42\x0c\n\ndeploymentB\x05\n\x03_idB\x10\n\x0e_error_messageB\x0f\n\r_created_timeB\x18\n\x16_start_processing_timeB\x11\n\x0f_completed_timeB\x0b\n\t_durationB\x19\n\x17_input_schema_ipfs_hash\"5\n\x0eModuleRunCheck\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\"H\n\x10\x43heckUserRequest\x12\x14\n\x07user_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\npublic_key\x18\x02 \x01(\tB\n\n\x08_user_id\"J\n\x11\x43heckUserResponse\x12\x15\n\ris_registered\x18\x01 \x01(\x08\x12\n\n\x02id\x18\x02 \x01(\t\x12\x12\n\npublic_key\x18\x03 \x01(\t\")\n\x13RegisterUserRequest\x12\x12\n\npublic_key\x18\x01 \x01(\t\"6\n\x14RegisterUserResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\npublic_key\x18\x02 \x01(\t2\x8a\x03\n\nGrpcServer\x12<\n\x08is_alive\x12\x16.google.protobuf.Empty\x1a\x16.agent.GeneralResponse\"\x00\x12\x38\n\x04stop\x12\x16.google.protobuf.Empty\x1a\x16.agent.GeneralResponse\"\x00\x12@\n\tCheckUser\x12\x17.agent.CheckUserRequest\x1a\x18.agent.CheckUserResponse\"\x00\x12I\n\x0cRegisterUser\x12\x1a.agent.RegisterUserRequest\x1a\x1b.agent.RegisterUserResponse\"\x00\x12:\n\tRunModule\x12\x17.agent.ModuleRunRequest\x1a\x10.agent.ModuleRun\"\x00\x30\x01\x12;\n\x0e\x43heckModuleRun\x12\x15.agent.ModuleRunCheck\x1a\x10.agent.ModuleRun\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_server_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GENERALRESPONSE']._serialized_start=87
  _globals['_GENERALRESPONSE']._serialized_end=133
  _globals['_LLMCONFIG']._serialized_start=136
  _globals['_LLMCONFIG']._serialized_end=369
  _globals['_NODESERVER']._serialized_start=371
  _globals['_NODESERVER']._serialized_end=435
  _globals['_NODECONFIGINPUT']._serialized_start=437
  _globals['_NODECONFIGINPUT']._serialized_end=546
  _globals['_NODECONFIG']._serialized_start=549
  _globals['_NODECONFIG']._serialized_end=1004
  _globals['_MODULE']._serialized_start=1007
  _globals['_MODULE']._serialized_end=1290
  _globals['_DATAGENERATIONCONFIG']._serialized_start=1293
  _globals['_DATAGENERATIONCONFIG']._serialized_end=1629
  _globals['_BASEDEPLOYMENT']._serialized_start=1632
  _globals['_BASEDEPLOYMENT']._serialized_end=1851
  _globals['_TOOLDEPLOYMENT']._serialized_start=1854
  _globals['_TOOLDEPLOYMENT']._serialized_end=2166
  _globals['_AGENTDEPLOYMENT']._serialized_start=2169
  _globals['_AGENTDEPLOYMENT']._serialized_end=2634
  _globals['_MODULERUNREQUEST']._serialized_start=2637
  _globals['_MODULERUNREQUEST']._serialized_end=3056
  _globals['_MODULERUN']._serialized_start=3059
  _globals['_MODULERUN']._serialized_end=3843
  _globals['_MODULERUNCHECK']._serialized_start=3845
  _globals['_MODULERUNCHECK']._serialized_end=3898
  _globals['_CHECKUSERREQUEST']._serialized_start=3900
  _globals['_CHECKUSERREQUEST']._serialized_end=3972
  _globals['_CHECKUSERRESPONSE']._serialized_start=3974
  _globals['_CHECKUSERRESPONSE']._serialized_end=4048
  _globals['_REGISTERUSERREQUEST']._serialized_start=4050
  _globals['_REGISTERUSERREQUEST']._serialized_end=4091
  _globals['_REGISTERUSERRESPONSE']._serialized_start=4093
  _globals['_REGISTERUSERRESPONSE']._serialized_end=4147
  _globals['_GRPCSERVER']._serialized_start=4150
  _globals['_GRPCSERVER']._serialized_end=4544
# @@protoc_insertion_point(module_scope)



from enum import Enum
from typing import Dict, List, Optional, Union, Any
from datetime import datetime
from pydantic import BaseModel, Field
from naptha_sdk.storage.schemas import StorageType

class User(BaseModel):
    id: str

class NodeServer(BaseModel):
    server_type: str
    port: int
    node_id: str

class NodeConfig(BaseModel):
    id: str
    owner: str
    public_key: str
    ip: str = Field(default="localhost")
    server_type: str = Field(default="ws")
    http_port: int = Field(default=7001)
    num_servers: int = Field(default=1)
    provider_types: List[str] = Field(default=["models", "storage", "modules"])
    servers: List[NodeServer]
    models: List[str]
    docker_jobs: bool
    ports: Optional[List[int]] = None
    routing_type: Optional[str] = Field(default="direct")
    routing_url: Optional[str] = Field(default=None)
    num_gpus: Optional[int] = Field(default=None)
    arch: Optional[str] = Field(default=None)
    os: Optional[str] = Field(default=None)
    ram: Optional[int] = Field(default=None)
    vram: Optional[int] = Field(default=None)

    class Config:
        allow_mutation = True

class NodeConfigUser(BaseModel):
    ip: str
    http_port: Optional[int] = None
    server_type: Optional[str] = None

class LLMClientType(str, Enum):
    OPENAI = "openai"
    AZURE_OPENAI = "azure_openai"
    ANTHROPIC = "anthropic"
    VLLM = "vllm"
    LITELLM = "litellm"
    OLLAMA = "ollama"
    STABILITY = "stability"

class LLMConfig(BaseModel):
    config_name: Optional[str] = "llm_config"
    client: Optional[LLMClientType] = None
    model: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    api_base: Optional[str] = None

class AgentModuleType(str, Enum):
    package = "package"
    docker = "docker"

class AgentConfig(BaseModel):
    config_name: Optional[str] = "agent_config"
    llm_config: Optional[LLMConfig] = None
    persona_module: Optional[Union[Dict, BaseModel]] = None
    system_prompt: Optional[Union[Dict, BaseModel]] = None

class ToolConfig(BaseModel):
    config_name: Optional[str] = None
    llm_config: Optional[LLMConfig] = None

class OrchestratorConfig(BaseModel):
    config_name: Optional[str] = "orchestrator_config"
    max_rounds: Optional[int] = 5

class EnvironmentConfig(BaseModel):
    config_name: Optional[str] = "environment_config"
    environment_type: Optional[str] = None

class KBConfig(BaseModel):
    config_name: Optional[str] = None
    storage_type: StorageType
    path: str
    schema: Dict[str, Any]
    options: Optional[Dict[str, Any]] = None

    def model_dict(self):
        if isinstance(self.storage_type, StorageType):
            self.storage_type = self.storage_type.value
        model_dict = self.dict()
        model_dict['storage_type'] = self.storage_type
        return model_dict

class DataGenerationConfig(BaseModel):
    save_outputs: Optional[bool] = None
    save_outputs_location: Optional[str] = None
    save_outputs_path: Optional[str] = None
    save_inputs: Optional[bool] = None
    save_inputs_location: Optional[str] = None
    default_filename: Optional[str] = None

class ToolDeployment(BaseModel):
    node: Union[NodeConfigUser, NodeConfig, Dict]
    name: Optional[str] = None
    module: Optional[Dict] = None
    config: Optional[ToolConfig] = None
    data_generation_config: Optional[DataGenerationConfig] = None

class KBDeployment(BaseModel):
    node: Union[NodeConfigUser, NodeConfig, Dict]
    name: Optional[str] = None
    module: Optional[Dict] = None
    config: Optional[KBConfig] = None

    def model_dict(self):
        model_dict = self.dict()
        if isinstance(self.config, KBConfig):
            model_dict['config'] = self.config.model_dict()
        return model_dict

class EnvironmentDeployment(BaseModel):
    node: Union[NodeConfigUser, NodeConfig, Dict]
    name: Optional[str] = None
    module: Optional[Dict] = None
    config: Optional[EnvironmentConfig] = None

class AgentDeployment(BaseModel):
    node: Union[NodeConfigUser, NodeConfig, Dict]
    name: Optional[str] = None
    module: Optional[Dict] = None
    config: Optional[AgentConfig] = None
    data_generation_config: Optional[DataGenerationConfig] = None
    tool_deployments: Optional[List[ToolDeployment]] = None
    environment_deployments: Optional[List[EnvironmentDeployment]] = None
    kb_deployments: Optional[List[KBDeployment]] = None

class OrchestratorDeployment(BaseModel):
    node: Union[NodeConfigUser, NodeConfig, Dict]
    name: Optional[str] = None
    module: Optional[Dict] = None
    config: Optional[OrchestratorConfig] = None
    agent_deployments: Optional[List[AgentDeployment]] = None
    environment_deployments: Optional[List[EnvironmentDeployment]] = None
    kb_deployments: Optional[List[KBDeployment]] = None

class DockerParams(BaseModel):
    docker_image: str
    docker_command: Optional[str] = ""
    docker_num_gpus: Optional[int] = 0
    docker_env_vars: Optional[Dict] = None
    input_dir: Optional[str] = None
    input_ipfs_hash: Optional[str] = None
    docker_input_dir: Optional[str] = None
    docker_output_dir: Optional[str] = None
    save_location: str = "node"

    class Config:
        allow_mutation = True

    class Config:
        allow_mutation = True
        json_encoders = {
            datetime: lambda v: v.isoformat(),
        }

    def model_dict(self):
        model_dict = self.dict()
        for key, value in model_dict.items():
            if isinstance(value, datetime):
                model_dict[key] = value.isoformat()
        return model_dict

class AgentRun(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: AgentDeployment
    tool_deployments: Optional[List[ToolDeployment]] = None
    environment_deployments: Optional[List[EnvironmentDeployment]] = None
    orchestrator_runs: List['OrchestratorRun'] = []
    status: str = "pending"
    error: bool = False
    id: Optional[str] = None
    results: list[str] = []
    error_message: Optional[str] = None
    created_time: Optional[str] = None
    start_processing_time: Optional[str] = None
    completed_time: Optional[str] = None
    duration: Optional[float] = None
    input_schema_ipfs_hash: Optional[str] = None
    signature: str
    class Config:
        allow_mutation = True
        json_encoders = {
            datetime: lambda v: v.isoformat(),
        }

    def model_dict(self):
        model_dict = self.dict()
        for key, value in model_dict.items():
            if isinstance(value, datetime):
                model_dict[key] = value.isoformat()
            elif isinstance(value, AgentModuleType):
                model_dict[key] = value.value
        for i, orchestrator_run in enumerate(model_dict['orchestrator_runs']):
            for key, value in orchestrator_run.items():
                if isinstance(value, datetime):
                    model_dict['orchestrator_runs'][i][key] = value.isoformat()
        return model_dict

class AgentRunInput(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: AgentDeployment
    tool_deployments: Optional[List[ToolDeployment]] = None
    environment_deployments: Optional[List[EnvironmentDeployment]] = None
    kb_deployment: Optional[KBDeployment] = None
    orchestrator_runs: List['OrchestratorRun'] = []
    signature: str

    def model_dict(self):
        model_dict = self.dict()
        if isinstance(self.deployment.config, BaseModel):
            config = self.deployment.config.model_dump()
            model_dict['deployment']['config'] = config
        return model_dict


class ToolRunInput(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: ToolDeployment
    agent_run: Optional[AgentRun] = None
    signature: str

    def model_dict(self):
        if isinstance(self.inputs, BaseModel):
            self.inputs = self.inputs.model_dump()
        model_dict = self.dict()
        model_dict['inputs'] = self.inputs
        return model_dict

class ToolRun(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: ToolDeployment
    agent_run: Optional[AgentRun] = None
    status: str = "pending"
    error: bool = False
    id: Optional[str] = None
    results: list[str] = []
    error_message: Optional[str] = None
    created_time: Optional[str] = None
    start_processing_time: Optional[str] = None
    completed_time: Optional[str] = None
    duration: Optional[float] = None
    signature: str
class OrchestratorRunInput(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: OrchestratorDeployment
    signature: str

    def model_dict(self):
        model_dict = self.dict()
        return model_dict

class OrchestratorRun(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: OrchestratorDeployment
    status: str = "pending"
    error: bool = False
    id: Optional[str] = None
    results: list[str] = []
    error_message: Optional[str] = None
    created_time: Optional[str] = None
    start_processing_time: Optional[str] = None
    completed_time: Optional[str] = None
    duration: Optional[float] = None
    agent_runs: List['AgentRun'] = []
    input_schema_ipfs_hash: Optional[str] = None
    signature: str

class EnvironmentRunInput(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: EnvironmentDeployment
    orchestrator_runs: List['OrchestratorRun'] = []
    signature: str

    def model_dict(self):
        model_dict = self.dict()
        return model_dict

class EnvironmentRun(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: EnvironmentDeployment
    orchestrator_runs: List['OrchestratorRun'] = []
    status: str = "pending"
    error: bool = False
    id: Optional[str] = None
    results: list[str] = []
    error_message: Optional[str] = None
    created_time: Optional[str] = None
    start_processing_time: Optional[str] = None
    completed_time: Optional[str] = None
    duration: Optional[float] = None
    input_schema_ipfs_hash: Optional[str] = None
    signature: str

class KBRunInput(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: KBDeployment
    orchestrator_runs: List['OrchestratorRun'] = []
    signature: str

    def model_dict(self):
        model_dict = self.dict()
        if isinstance(self.deployment, KBDeployment):
            model_dict['deployment'] = self.deployment.model_dict()
        return model_dict

class KBRun(BaseModel):
    consumer_id: str
    inputs: Optional[Union[Dict, BaseModel, DockerParams]] = None
    deployment: KBDeployment
    orchestrator_runs: List['OrchestratorRun'] = []
    status: str = "pending"
    error: bool = False
    id: Optional[str] = None
    results: list[Optional[str]] = []
    error_message: Optional[str] = None
    created_time: Optional[str] = None
    start_processing_time: Optional[str] = None
    completed_time: Optional[str] = None
    duration: Optional[float] = None
    signature: str

class ChatMessage(BaseModel):
    role: str
    content: str

class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatMessage]
    temperature: Optional[float] = None
    max_tokens: Optional[int] = None
    top_p: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    stop: Optional[List[str]] = None
    stream: Optional[bool] = None
    stream_options: Optional[dict] = None
    n: Optional[int] = None
    response_format: Optional[dict] = None
    seed: Optional[int] = None
    tools: Optional[List] = None
    tool_choice: Optional[str] = None
    parallel_tool_calls: Optional[bool] = None

class Choices(BaseModel):
    message: ChatMessage
    finish_reason: str
    index: int

class ModelResponse(BaseModel):
    id: str
    choices: List[Choices]
    created: int
    model: str
    object: str



from enum import Enum
from pydantic import BaseModel, Field
from typing import Dict, Any, Optional, Union, List, Literal, BinaryIO
from io import BufferedReader, BytesIO, IOBase

class StorageType(str, Enum):
    DATABASE = "db"
    FILESYSTEM = "fs"
    IPFS = "ipfs"

class StorageLocation(BaseModel):
    storage_type: StorageType
    path: str
    options: Dict[str, Any] = Field(default_factory=dict)

    @property
    def uri(self) -> str:
        """Get URI representation of location"""
        return f"{self.storage_type.value}://{self.path}"
    
    @classmethod
    def from_uri(cls, uri: str) -> "StorageLocation":
        """Create StorageLocation from URI string"""
        scheme, path = uri.split("://", 1)
        return cls(storage_type=StorageType(scheme), path=path)

class StorageMetadata(BaseModel):
    content_type: Optional[str] = None
    created_at: Optional[str] = None
    modified_at: Optional[str] = None
    size: Optional[int] = None
    checksum: Optional[str] = None
    tags: Dict[str, str] = Field(default_factory=dict)
    custom: Dict[str, Any] = Field(default_factory=dict)

class StorageObject(BaseModel):
    location: StorageLocation
    data: Optional[Any] = None
    metadata: StorageMetadata = Field(default_factory=StorageMetadata)

class DatabaseReadOptions(BaseModel):
    """Options specific to database reads"""
    columns: Optional[List[str]] = None
    conditions: Optional[List[Dict[str, Any]]] = None
    order_by: Optional[str] = None
    order_direction: Optional[str] = "asc"
    limit: Optional[int] = None
    offset: Optional[int] = None
    # Added fields for QA/vector search
    query_vector: Optional[List[float]] = None
    query_col: Optional[str] = None  # Column to search against
    answer_col: Optional[str] = None  # Column to return as answer
    vector_col: Optional[str] = None  # Column containing vectors
    top_k: Optional[int] = Field(default=5, ge=1)  # Number of results for vector search
    include_similarity: Optional[bool] = Field(default=True)  # Include similarity scores

class IPFSOptions(BaseModel):
    """Options specific to IPFS operations"""
    # IPNS options
    ipns_operation: Optional[Literal["create", "update", "none"]] = Field(
        default="none", 
        description="Whether to create new IPNS record, update existing, or skip IPNS"
    )
    ipns_name: Optional[str] = Field(
        None, 
        description="IPNS name to update when ipns_operation is 'update'"
    )
    
    # Pinning options
    unpin_previous: bool = Field(
        default=False, 
        description="Unpin previous hash when updating"
    )
    previous_hash: Optional[str] = Field(
        None,
        description="Previous IPFS hash to unpin when unpin_previous is True"
    )
    
    # Read options 
    resolve_ipns: bool = Field(
        default=False, 
        description="Resolve IPNS name to IPFS hash"
    )

class StorageRequestType(str, Enum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"
    LIST = "list"
    SEARCH = "search"

class BaseStorageRequest(BaseModel):
    request_type: StorageRequestType
    storage_type: StorageType
    path: str
    options: Union[Dict[str, Any], DatabaseReadOptions] = Field(default_factory=dict)

    def model_dict(self):
        model_dict = self.dict()
        if isinstance(self.options, BaseModel):
            options = self.options.model_dump()
            model_dict['options'] = options
        model_dict['storage_type'] = self.storage_type.value
        model_dict['request_type'] = self.request_type.value
        return model_dict

    class Config:
        arbitrary_types_allowed = True

class CreateStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.CREATE, literal=True)
    data: Optional[Dict[str, Any]] = None
    file: Optional[IOBase] = None 

class ReadStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.READ, literal=True)

class UpdateStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.UPDATE, literal=True)
    data: Union[Dict[str, Any], bytes, BinaryIO]

class DeleteStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.DELETE, literal=True)
    condition: Optional[Dict[str, Any]] = None

class ListStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.LIST, literal=True)

class SearchStorageRequest(BaseStorageRequest):
    request_type: StorageRequestType = Field(default=StorageRequestType.SEARCH, literal=True)
    query: Any
    query_type: str = "text"
    limit: Optional[int] = None




from dotenv import load_dotenv
import os
from naptha_sdk.utils import add_credentials_to_env, get_logger, write_private_key_to_file
from naptha_sdk.user import generate_keypair
from naptha_sdk.user import get_public_key, is_hex
from surrealdb import Surreal
import traceback
from typing import Dict, List, Optional, Tuple

import jwt
from surrealdb import Surreal

from naptha_sdk.user import generate_keypair
from naptha_sdk.user import get_public_key
from naptha_sdk.utils import add_credentials_to_env, get_logger

logger = get_logger(__name__)

load_dotenv()

class Hub:
    """The Hub class is the entry point into Naptha AI Hub."""

    def __init__(self, hub_url, public_key=None, *args, **kwargs):
        self.hub_url = hub_url
        self.public_key = public_key
        self.ns = "naptha"
        self.db = "naptha"
        self.surrealdb = Surreal(hub_url)
        self.is_authenticated = False
        self.user_id = None
        self.token = None
        
        logger.info(f"Hub URL: {hub_url}")

    async def connect(self):
        """Connect to the database and authenticate"""
        if not self.is_authenticated:
            try:
                await self.surrealdb.connect()
                await self.surrealdb.use(namespace=self.ns, database=self.db)
            except Exception as e:
                logger.error(f"Connection failed: {e}")
                raise

    def _decode_token(self, token: str) -> str:
        return jwt.decode(token, options={"verify_signature": False})["ID"]

    async def signin(self, username: str, password: str) -> Tuple[bool, Optional[str], Optional[str]]:
        logger.info(f"Attempting authentication for user: {username}")
        
        try:
            user = await self.surrealdb.signin({
                "NS": self.ns,
                "DB": self.db,
                "AC": "user",
                "username": username,
                "password": password,
            })
            
            self.user_id = self._decode_token(user)
            local_public_key = self.public_key or get_public_key(os.getenv("PRIVATE_KEY"))
            hub_public_key = self.user_id.split(":")[1]
            
            if local_public_key != hub_public_key:
                logger.error("Public key mismatch", extra={
                    "local_key": local_public_key,
                    "hub_key": hub_public_key
                })
                raise Exception(
                    "Public key mismatch. Please verify your private key in .env file. "
                    f"Local key: {local_public_key}, Hub key: {hub_public_key}"
                )
            
            self.token = user
            self.is_authenticated = True
            
            logger.info(f"Successfully authenticated user: {username}")
            return True, user, self.user_id
            
        except Exception as e:
            raise Exception(f"Authentication failed: {str(e)}. Please ensure you have set HUB_URL, HUB_USERNAME, and HUB_PASSWORD in the .env file, and that you have run `naptha signup` for this user.")
    
    async def signup(self, username: str, password: str, public_key: str) -> Tuple[bool, Optional[str], Optional[str]]:
        user = await self.surrealdb.signup({
            "NS": self.ns,
            "DB": self.db,
            "AC": "user",
            "name": username,
            "username": username,
            "password": password,
            "public_key": public_key,
        })
        if not user:
            return False, None, None
        self.user_id = self._decode_token(user)
        return True, user, self.user_id


    async def get_user(self, user_id: str) -> Optional[Dict]:
        return await self.surrealdb.select(user_id)

    async def get_user_by_username(self, username: str) -> Optional[Dict]:
        result = await self.surrealdb.query(
            "SELECT * FROM user WHERE username = $username LIMIT 1",
            {"username": username}
        )
        
        if result and result[0]["result"]:
            return result[0]["result"][0]
        
        return None

    async def get_user_by_public_key(self, public_key: str) -> Optional[Dict]:
        result = await self.surrealdb.query(
            "SELECT * FROM user WHERE public_key = $public_key LIMIT 1",
            {"public_key": public_key}
        )

        if result and result[0]["result"]:
            return result[0]["result"][0]
        return None

    async def get_credits(self) -> List:
        user = await self.get_user(self.user_id)
        return user['credits']

    async def get_node(self, node_id: str) -> Optional[Dict]:
        return await self.surrealdb.select(node_id)

    async def list_servers(self) -> List:
        servers = await self.surrealdb.query("SELECT * FROM server;")
        return servers[0]['result']

    async def list_nodes(self, node_ip=None) -> List:
        if not node_ip:
            nodes = await self.surrealdb.query("SELECT * FROM node;")
            return nodes[0]['result']
        else:
            nodes = await self.surrealdb.query("SELECT * FROM node WHERE ip=$node_ip;", {"node_ip": node_ip})
            node = nodes[0]['result'][0]
            server_ids = node['servers']
            servers = []
            for server_id in server_ids:
                server = await self.surrealdb.select(server_id)
                servers.append(server)
            node['servers'] = servers

            alt_ports = [
                server['port'] 
                for server in servers
                if server['server_type'] in ['ws', 'grpc']
            ]
            node['ports'] = alt_ports
            return node

    async def list_agents(self, agent_name=None) -> List:
        if not agent_name:
            agents = await self.surrealdb.query("SELECT * FROM agent;")
            return agents[0]['result']
        else:
            agent = await self.surrealdb.query("SELECT * FROM agent WHERE id=$agent_name;", {"agent_name": agent_name})
            return agent[0]['result']

    async def list_tools(self, tool_name=None) -> List:
        if not tool_name:
            tools = await self.surrealdb.query("SELECT * FROM tool;")
            return tools[0]['result']
        else:
            tool = await self.surrealdb.query("SELECT * FROM tool WHERE name=$tool_name;", {"tool_name": tool_name})
            return tool[0]['result']

    async def list_orchestrators(self, orchestrator_name=None) -> List:
        if not orchestrator_name:
            orchestrators = await self.surrealdb.query("SELECT * FROM orchestrator;")
            return orchestrators[0]['result']
        else:
            orchestrator = await self.surrealdb.query("SELECT * FROM orchestrator WHERE id=$orchestrator_name;", {"orchestrator_name": orchestrator_name})
            return orchestrator[0]['result']

    async def list_environments(self, environment_name=None) -> List:
        if not environment_name:
            environments = await self.surrealdb.query("SELECT * FROM environment;")
            return environments[0]['result']
        else:
            environment = await self.surrealdb.query("SELECT * FROM environment WHERE id=$environment_name;", {"environment_name": environment_name})
            return environment[0]['result']

    async def list_personas(self, persona_name=None) -> List:
        if not persona_name:
            personas = await self.surrealdb.query("SELECT * FROM persona;")
            return personas[0]['result']
        else:
            if not "persona:" in persona_name:
                persona_name = f"persona:{persona_name}"
            persona = await self.surrealdb.query("SELECT * FROM persona WHERE id=$persona_name;", {"persona_name": persona_name})
            return persona[0]['result']
    
    async def list_memories(self, memory_name=None) -> List:
        if not memory_name:
            memories = await self.surrealdb.query("SELECT * FROM memory;")
            return memories[0]['result']
        else:
            memory = await self.surrealdb.query("SELECT * FROM memory WHERE id=$memory_name;", {"memory_name": memory_name})
            return memory[0]['result']

    async def list_kbs(self, kb_name=None) -> List:
        if not kb_name:
            kbs = await self.surrealdb.query("SELECT * FROM kb;")
            return kbs[0]['result']
        else:
            kb = await self.surrealdb.query("SELECT * FROM kb WHERE name=$kb_name;", {"kb_name": kb_name})
            return kb[0]['result']

    async def list_modules(self, module_type, module_name) -> List:
        if module_type == 'agent':
            modules = await self.surrealdb.query("SELECT * FROM agent WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'tool':
            modules = await self.surrealdb.query("SELECT * FROM tool WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'orchestrator':
            modules = await self.surrealdb.query("SELECT * FROM orchestrator WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'environment':
            modules = await self.surrealdb.query("SELECT * FROM environment WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'persona':
            modules = await self.surrealdb.query("SELECT * FROM persona WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'memory':
            modules = await self.surrealdb.query("SELECT * FROM memory WHERE id=$module_name;", {"module_name": module_name})
        elif module_type == 'kb':
            modules = await self.surrealdb.query("SELECT * FROM kb WHERE id=$module_name;", {"module_name": module_name})
        return modules[0]['result']

    async def list_kb_content(self, kb_name: str) -> List:
        kb_content = await self.surrealdb.query("SELECT * FROM kb_content WHERE kb_id=$kb_id;", {"kb_id": f"kb:{kb_name}"})
        return kb_content[0]['result']

    async def delete_agent(self, agent_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in agent_id:
            agent_id = f"agent:{agent_id}".strip()
        print(f"Deleting agent: {agent_id}")
        success = await self.surrealdb.delete(agent_id)
        if success:
            print("Deleted agent")
        else:
            print("Failed to delete agent")
        return success

    async def delete_tool(self, tool_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in tool_id:
            tool_id = f"tool:{tool_id}".strip()
        print(f"Deleting tool: {tool_id}")
        success = await self.surrealdb.delete(tool_id)
        if success:
            print("Deleted tool")
        else:
            print("Failed to delete tool")
        return success

    async def delete_orchestrator(self, orchestrator_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in orchestrator_id:
            orchestrator_id = f"orchestrator:{orchestrator_id}".strip()
        print(f"Deleting orchestrator: {orchestrator_id}")
        success = await self.surrealdb.delete(orchestrator_id)
        if success:
            print("Deleted orchestrator")
        else:
            print("Failed to delete orchestrator")
        return success

    async def delete_environment(self, environment_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in environment_id:
            environment_id = f"environment:{environment_id}".strip()
        print(f"Deleting environment: {environment_id}")
        success = await self.surrealdb.delete(environment_id)
        if success:
            print("Deleted environment")
        else:
            print("Failed to delete environment")
        return success

    async def delete_persona(self, persona_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in persona_id:
            persona_id = f"persona:{persona_id}".strip()
        print(f"Deleting persona: {persona_id}")
        success = await self.surrealdb.delete(persona_id)
        if success:
            print("Deleted persona")
        else:
            print("Failed to delete persona")
        return success
    
    async def delete_memory(self, memory_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in memory_id:
            memory_id = f"memory:{memory_id}".strip()
        print(f"Deleting memory: {memory_id}")
        success = await self.surrealdb.delete(memory_id)
        if success:
            print("Deleted memory")
        else:
            print("Failed to delete memory")
        return success

    async def delete_kb(self, kb_id: str) -> Tuple[bool, Optional[Dict]]:
        if ":" not in kb_id:
            kb_id = f"kb:{kb_id}".strip()
        print(f"Deleting knowledge base: {kb_id}")
        success = await self.surrealdb.delete(kb_id)
        if success:
            print("Deleted knowledge base")
        else:
            print("Failed to delete knowledge base")
        return success

    async def create_agent(self, agent_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not agent_config.get('id'):
            return await self.surrealdb.create("agent", agent_config)
        else:
            return await self.surrealdb.create(agent_config.pop('id'), agent_config)

    async def create_tool(self, tool_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not tool_config.get('id'):
            return await self.surrealdb.create("tool", tool_config)
        else:
            return await self.surrealdb.create(tool_config.pop('id'), tool_config)

    async def create_orchestrator(self, orchestrator_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not orchestrator_config.get('id'):
            return await self.surrealdb.create("orchestrator", orchestrator_config)
        else:
            return await self.surrealdb.create(orchestrator_config.pop('id'), orchestrator_config)

    async def create_environment(self, environment_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not environment_config.get('id'):
            return await self.surrealdb.create("environment", environment_config)
        else:
            return await self.surrealdb.create(environment_config.pop('id'), environment_config)

    async def create_persona(self, persona_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not persona_config.get('id'):
            return await self.surrealdb.create("persona", persona_config)
        else:
            return await self.surrealdb.create(persona_config.pop('id'), persona_config)
        
    async def create_memory(self, memory_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not memory_config.get('id'):
            return await self.surrealdb.create("memory", memory_config)
        else:
            return await self.surrealdb.create(memory_config.pop('id'), memory_config)

    async def create_kb(self, kb_config: Dict) -> Tuple[bool, Optional[Dict]]:
        if not kb_config.get('id'):
            return await self.surrealdb.create("kb", kb_config)
        else:
            return await self.surrealdb.create(kb_config.pop('id'), kb_config)

    async def update_agent(self, agent_config: Dict) -> Tuple[bool, Optional[Dict]]:
        return await self.surrealdb.update("agent", agent_config)

    async def create_or_update_module(self, module_type, module_config: Dict) -> Tuple[bool, Optional[Dict]]:
        list_modules = await self.list_modules(module_type, module_config.get('id'))
        if not list_modules:
            logger.info(f"Module does not exist. Registering new module: {module_config.get('id')}")
            return await self.surrealdb.create(module_type, module_config)
        else:
            logger.info(f"Module already exists. Updating existing module: {module_config.get('id')}")
            return await self.surrealdb.update(module_config.pop('id'), module_config)

    async def close(self):
        """Close the database connection"""
        try:
            await self.surrealdb.close()
        except Exception as e:
            logger.error(f"Error closing database connection: {e}")
        finally:
            self.is_authenticated = False
            self.user_id = None
            self.token = None

    async def __aenter__(self):
        """Async enter method for context manager"""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async exit method for context manager"""
        await self.close()


async def user_setup_flow(hub_url, public_key):
    async with Hub(hub_url, public_key) as hub:
        username, password = os.getenv("HUB_USERNAME"), os.getenv("HUB_PASSWORD")
        username_exists, password_exists = len(username) > 1, len(password) > 1
        public_key = get_public_key(os.getenv("PRIVATE_KEY")) if os.getenv("PRIVATE_KEY") else None
        logger.info(f"Checking if user exists... User: {username}")
        user = await hub.get_user_by_username(username)
        user_public_key = await hub.get_user_by_public_key(public_key)
        existing_public_key_user = user_public_key is not None and user_public_key.get('username') != username

        match user, username_exists, password_exists, existing_public_key_user:
            case _, True, _, True:
                # Public key exists and doesn't match the provided username and password
                raise Exception(f"Using user credentials in .env. User with public key {public_key} already exists but doesn't match the provided username and password. Please use a different private key (or set blank in the .env file to randomly generate with launch.sh).")

            case _, False, False, True:
                # Public key exists and no username/password provided
                raise Exception(f"Using private key in .env. User with public key {public_key} already exists. Cannot create new user. Please use a different private key (or set blank in the .env file to randomly generate with launch.sh).")

            case None, False, _, False:
                # User doesn't exist and credentials are missing
                logger.info("No credentials provided in .env.")
                create_new = input("Would you like to create a new user? (yes/no): ").lower()
                if create_new != 'yes':
                    raise Exception("User does not exist and new user creation was declined.")
                logger.info("Creating new user...")
                while True:
                    username = input("Enter username: ")
                    existing_user = await hub.get_user_by_username(username)
                    if existing_user:
                        print(f"Username '{username}' already exists. Please choose a different username.")
                        continue
                    password = input("Enter password: ")
                    public_key, private_key_path = generate_keypair(f"{username}.pem")
                    print(f"Signing up user: {username} with public key: {public_key}")
                    success, token, user_id = await hub.signup(username, password, public_key)
                    if success:
                        add_credentials_to_env(username, password, private_key_path)
                        logger.info("Sign up successful!")
                        return token, user_id
                    else:
                        logger.error("Sign up failed. Please try again.")

            case None, True, True, False:
                logger.info("User doesn't exist but some credentials are provided in .env. Using them to create new user.")
                private_key_path = None
                if not public_key:
                    logger.info("No public key provided. Generating new keypair...")
                    public_key, private_key_path = generate_keypair(f"{username}.pem")

                print(f"Signing up user: {username} with public key: {public_key}")
                success, token, user_id = await hub.signup(username, password, public_key)
                if success:
                    if private_key_path:
                        add_credentials_to_env(username, password, private_key_path)
                    logger.info("Sign up successful!")
                    return token, user_id
                else:
                    logger.error("Sign up failed.")
                    raise Exception("Sign up failed.")

            case dict(), True, True, False:
                # User exists, attempt to sign in
                logger.info("Using user credentials in .env. User exists. Attempting to sign in...")
                if os.getenv("PRIVATE_KEY") and is_hex(os.getenv("PRIVATE_KEY")):
                    write_private_key_to_file(os.getenv("PRIVATE_KEY"), username)

                success, token, user_id = await hub.signin(username, password)
                if success:
                    logger.info("Sign in successful!")
                    return token, user_id
                else:
                    logger.error("Sign in failed. Please check your credentials in the .env file.")
                    raise Exception("Sign in failed. Please check your credentials in the .env file.")

            case _:
                logger.error("Unexpected case encountered in user setup flow.")
                raise Exception("Unexpected error in user setup. Please check your configuration and try again.")
            
async def list_nodes(node_ip: str) -> List:

    hub_username = os.getenv("HUB_USERNAME")
    hub_password = os.getenv("HUB_PASSWORD")
    hub_url = os.getenv("HUB_URL")
    
    async with Hub(hub_url) as hub:
        try:
            _, _, _ = await hub.signin(hub_username, hub_password)
        except Exception as auth_error:
            raise ConnectionError(f"Failed to authenticate with Hub: {str(auth_error)}")

        node = await hub.list_nodes(node_ip=node_ip)
        return node


from naptha_sdk.client.node import NodeClient
from naptha_sdk.schemas import AgentRun, ToolRunInput
from naptha_sdk.utils import get_logger
from typing import Union
from dotenv import load_dotenv
import os

logger = get_logger(__name__)
load_dotenv(override=True)
class Tool:
    def __init__(self, 
        tool_deployment,
        *args,
        **kwargs
    ):
        self.tool_deployment = tool_deployment
        self.tool_node = NodeClient(self.tool_deployment.node)

    async def call_tool_func(self, module_run: Union[AgentRun, ToolRunInput]):
        logger.info(f"Running tool on worker node {self.tool_node}")
        tool_run = await self.tool_node.run_module(module_type="tool", run_input=module_run.model_dict())
        return tool_run


from ecdsa import SigningKey, SECP256k1
import os, re

def generate_keypair(private_key_filename=None):
    pkfile = private_key_filename if private_key_filename else 'private_key.pem'
    file_path = os.path.join(os.getcwd(), pkfile)
    
    # Check if private key file exists and is not empty
    if private_key_filename and os.path.isfile(file_path):
        with open(file_path, 'r') as file:
            private_key = file.read().strip()
            # Check if file is empty or contains only whitespace
            if not private_key:
                # Generate new key if file is empty
                private_key = SigningKey.generate(curve=SECP256k1).to_string().hex()
                with open(file_path, 'w') as f:
                    f.write(private_key)
    else:
        # Generate new private key
        private_key = SigningKey.generate(curve=SECP256k1).to_string().hex()
        # Save the private key to a file
        with open(file_path, 'w') as file:
            file.write(private_key)

    public_key = generate_public_key(private_key)
    return public_key, file_path

def get_public_key(private_key):
    # To ensure old users can still login
    if private_key and is_hex(private_key):
        private_key_hex = private_key
    elif private_key and os.path.isfile(private_key):
        with open(private_key) as file:
            content = file.read().strip()
            
            if content:
                private_key_hex = content
            else:
                return None
    else:
        return None

    return generate_public_key(private_key_hex)
    
def generate_public_key(private_key_hex):
    private_key = SigningKey.from_string(bytes.fromhex(private_key_hex), curve=SECP256k1)
    public_key = private_key.get_verifying_key()

    return public_key.to_string().hex()

def is_hex(string):
    # Check if the string matches the pattern for a hexadecimal key
    return bool(re.match(r'^[0-9a-fA-F]{64}$', string))

def sign_consumer_id(consumer_id, private_key):
    if private_key and is_hex(private_key):
        private_key_hex = private_key
    elif private_key and os.path.isfile(private_key):
        with open(private_key) as file:
            content = file.read().strip()
            
            if content:
                private_key_hex = content
            else:
                return None
    else:
        return None
    
    private_key = SigningKey.from_string(bytes.fromhex(private_key_hex), curve=SECP256k1)
    consumer_id_bytes = consumer_id.encode('utf-8')
    signature = private_key.sign(consumer_id_bytes)
    return signature.hex()


from dotenv import load_dotenv
from git import Repo
import importlib.util
import ipfshttpclient
import json
from naptha_sdk.client.hub import Hub
from naptha_sdk.utils import get_logger
import os
from pathlib import Path
import re
from pydantic import BaseModel
import subprocess
import tempfile
import textwrap
import tomlkit
import yaml
import zipfile
import fnmatch

load_dotenv()
logger = get_logger(__name__)

IPFS_GATEWAY_URL="/dns/provider.akash.pro/tcp/31832/http"
AGENT_DIR = "agent_pkgs"

# Certain packages cause issues with dependencies and can be slow to resolve, better to specify ranges
PACKAGE_VERSIONS = {
    "crewai": "^0.41.1",
    "crewai_tools": ">=0.4.6,<0.5.0",
    "embedchain": ">=0.1.113,<0.2.0",
}

def init_agent_package(package_name):
    subprocess.run(["poetry", "new", f"{AGENT_DIR}/{package_name}"])
    subprocess.run(["git", "init", f"{AGENT_DIR}/{package_name}"])

def is_std_lib(module_name):
    try:
        module_spec = importlib.util.find_spec(module_name)
        return module_spec is not None and 'site-packages' not in module_spec.origin
    except ImportError:
        return False

def add_dependencies_to_pyproject(package_name, packages):
    # Adds dependencies with wildcard versioning
    with open(f"{AGENT_DIR}/{package_name}/pyproject.toml", 'r', encoding='utf-8') as file:
        data = tomlkit.parse(file.read())

    dependencies = data['tool']['poetry']['dependencies']
    dependencies["python"] = ">=3.10,<3.13"
    dependencies["naptha-sdk"] = {
        "git": "https://github.com/NapthaAI/naptha-sdk.git",
        "branch": "feat/run-agent-tools"
    }

    packages_to_add = []
    for package in packages:
        curr_package = package['module'].split('.')[0]
        if curr_package not in packages_to_add and not is_std_lib(curr_package):
            dependencies[curr_package] = PACKAGE_VERSIONS.get(curr_package, "*")
    dependencies["python-dotenv"] = "*"

    # Serialize the TOML data and write it back to the file
    with open(f"{AGENT_DIR}/{package_name}/pyproject.toml", 'w', encoding='utf-8') as file:
        file.write(tomlkit.dumps(data))

def render_agent_code(agent_name, agent_code, obj_name, local_modules, selective_import_modules, standard_import_modules, variable_modules, union_modules, params):
    # Add the imports for installed modules (e.g. crewai)
    content = ''

    for module in standard_import_modules:
        line = f'import {module["name"]} \n'
        content += line

    for module in selective_import_modules:
        line = f'from {module["module"]} import {module["name"]} \n'
        content += line

    for module in variable_modules:
        if module["module"] and module["import_needed"]:
            content += f'from {module["module"]} import {module["name"]} \n'

    if any('crewai' in module['module'] for module in selective_import_modules):
        content += "from crewai import Task\n"

    for module in union_modules:
        content += module['source']

    # Add the naptha imports and logger setup
    naptha_imports = f'''from dotenv import load_dotenv
from {agent_name}.schemas import InputSchema
from naptha_sdk.utils import get_logger

logger = get_logger(__name__)

load_dotenv()

'''
    content += naptha_imports
    for module in selective_import_modules:
        if 'source' in module and module['source']:
            content += module['source'] + "\n"

    # Add the source code for the local modules 
    for module in local_modules:
        content += module['source'] + "\n"

    for module in variable_modules:
        content += module['source'] + "\n"

    # Convert class method to function
    agent_code = agent_code.replace('self.', '')
    agent_code = agent_code.replace('self', '')

    content += textwrap.dedent(agent_code) + "\n\n"

    param_str = ", ".join(f"inputs.{name}" for name, info in params.items())

    # Define the new function signature
    content += f"""def run(inputs: InputSchema, *args, **kwargs):
    {agent_name}_0 = {obj_name}({param_str})

    tool_input_class = globals().get(inputs.tool_input_type)
    tool_input = tool_input_class(**inputs.tool_input_value)
    method = getattr({agent_name}_0, inputs.tool_name, None)

    return method(tool_input)

if __name__ == "__main__":
    from naptha_sdk.utils import load_yaml
    from {agent_name}.schemas import InputSchema

    cfg_path = "{agent_name}/component.yaml"
    cfg = load_yaml(cfg_path)

    # You will likely need to change the inputs dict
    inputs = {{"tool_name": "execute_task", "tool_input_type": "Task", "tool_input_value": {{"description": "What is the market cap of AMZN?", "expected_output": "The market cap of AMZN"}}}}
    inputs = InputSchema(**inputs)

    response = run(inputs)
    print(response)
"""
    
    return content

def generate_component_yaml(agent_name, user_id):
    component = {
        'name': agent_name,
        'type': agent_name,
        'author': user_id,
        'version': '0.1.0',
        'description': agent_name,
        'license': 'MIT',
        'models': {
            'default_model_provider': 'ollama',
            'ollama': {
                'model': 'ollama/llama3.1:70b',
                'max_tokens': 1000,
                'temperature': 0,
                'api_base': 'http://localhost:11434'
            }
        },
        'inputs': {
            'system_message': 'You are a helpful AI assistant.',
            'save': False,
            'location': 'node'
        },
        'outputs': {
            'filename': 'output.txt',
            'save': False,
            'location': 'node'
        },
        'implementation': {
            'package': {
                'entrypoint': 'run.py'
            }
        }
    }

    with open(f'{AGENT_DIR}/{agent_name}/{agent_name}/component.yaml', 'w') as file:
        yaml.dump(component, file, default_flow_style=False)

def generate_schema(agent_name, params):
    schema_code = '''from pydantic import BaseModel
from typing import Any

class InputSchema(BaseModel):
    tool_name: str
    tool_input_type: str
    tool_input_value: dict
'''

    for name, info in params.items():
        print("INFO", name, info)
        if info['value'] is None:
            if 'List' in str(info['type']):
                schema_code += f'    {name}: list\n'
            elif info['type'] is None:
                schema_code += f'    {name}: Any\n'
            elif issubclass(info["type"], BaseModel):
                schema_code += f'    {name}: dict\n'
            else:
                schema_code += f'    {name}: {info["type"].__name__}\n'
        else:
            if 'List' in str(info['type']):
                schema_code += f'    {name}: list = {info["value"]}\n'
            elif info['type'] is None:
                schema_code += f'    {name}: Any = {info["value"]}\n'
            elif issubclass(info["type"], BaseModel):
                schema_code += f'    {name}: dict = {info["value"]}\n'
            else:
                schema_code += f'    {name}: {info["type"].__name__} = {info["value"]}\n'

    with open(f'{AGENT_DIR}/{agent_name}/{agent_name}/schemas.py', 'w') as file:
        file.write(schema_code)

def git_add_commit(agent_name):
    subprocess.run(["git", "-C", f"{AGENT_DIR}/{agent_name}", "add", "-A"])
    subprocess.run(["git", "-C", f"{AGENT_DIR}/{agent_name}", "commit", "-m", "Initial commit"])
    subprocess.run(["git", "-C", f"{AGENT_DIR}/{agent_name}", "tag", "-f", "v0.1"])

def write_code_to_package(agent_name, code):
    package_path = f'{AGENT_DIR}/{agent_name}'
    code_path = os.path.join(package_path, agent_name, 'run.py')

    os.makedirs(os.path.dirname(code_path), exist_ok=True)
    with open(code_path, 'w') as file:
        file.write(code)

def add_files_to_package(agent_name, params, user_id):
    package_path = f'{AGENT_DIR}/{agent_name}'

    # Generate schema and component yaml
    generate_schema(agent_name, params)
    generate_component_yaml(agent_name, user_id)

    # Create .env.example file
    env_example_path = os.path.join(package_path, '.env.example')
    with open(env_example_path, 'w') as env_file:
        env_file.write('OPENAI_API_KEY=\n')

def zip_dir(directory_path: str) -> None:
    """
    Zip the specified directory and write it to a file on disk.
    """
    output_zip_file = f"{directory_path}.zip"
    with zipfile.ZipFile(output_zip_file, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for root, dirs, files in os.walk(directory_path):
            for file in files:
                file_path = os.path.join(root, file)
                zip_file.write(file_path, os.path.relpath(file_path, directory_path))
    print(f"Zipped directory '{directory_path}' to '{output_zip_file}'")
    return output_zip_file

async def write_to_ipfs(file_path):
    """Write a file to IPFS, optionally publish to IPNS or update an existing IPNS record."""
    try:
        logger.info(f"Writing file to IPFS: {file_path}")
        if not IPFS_GATEWAY_URL:
            return (500, {"message": "IPFS_GATEWAY_URL not found"})
        
        client = ipfshttpclient.connect(IPFS_GATEWAY_URL)
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tmpfile:
            with open(file_path, "rb") as f:
                content = f.read()            
            tmpfile.write(content)
            tmpfile_name = tmpfile.name
        
        result = client.add(tmpfile_name)
        client.pin.add(result["Hash"])
        os.unlink(tmpfile_name)
        
        ipfs_hash = result["Hash"]
        response = {
            "message": "File written and pinned to IPFS",
            "ipfs_hash": ipfs_hash,
        }

        return (201, response)
    except Exception as e:
        logger.error(f"Error writing file to IPFS: {e}")
        import traceback
        logger.error(f"Error writing file to IPFS: {e}")
        logger.error(f"Traceback: {traceback.format_exc()}")
        return (500, {"message": f"Error writing file to IPFS: {e}"})

async def publish_ipfs_package(agent_name, decorator = False):
    package_path = f"{AGENT_DIR}/{agent_name}"

    if not decorator:
        output_zip_file = zip_dir_with_gitignore(Path.cwd())
    else:
        output_zip_file = zip_dir(package_path)
    
    success, response = await write_to_ipfs(output_zip_file)
    logger.info(f"Response: {response}")
    return success, response

# Function to sort modules based on dependencies
def sort_modules(modules, dependencies):
    sorted_modules = []
    unsorted_modules = modules.copy()

    while unsorted_modules:
        for mod in unsorted_modules:
            mod_deps = dependencies[mod['name']]
            if all(dep in [m['name'] for m in sorted_modules] for dep in mod_deps):
                sorted_modules.append(mod)
                unsorted_modules.remove(mod)
                break

    return sorted_modules

# Define a function to extract dependencies from the source code
def extract_dependencies(module, modules):
    dependencies = []
    for mod in modules:
        if mod['name'] != module['name']:
            # Use a negative lookahead to exclude matches within quotes
            pattern = r'\b' + re.escape(mod['name']) + r'\b(?=([^"\']*["\'][^"\']*["\'])*[^"\']*$)'
            if re.search(pattern, module['source']):
                dependencies.append(mod['name'])
    return dependencies

def load_input_schema(repo_name):
    """Loads the input schema"""
    schemas_module = importlib.import_module(f"{repo_name}.schemas")
    input_schema = getattr(schemas_module, "Persona")
    return input_schema

async def load_persona(persona_module):
    """Load persona from a JSON or YAML file in a git repository."""

    hub_username = os.getenv("HUB_USERNAME")
    hub_password = os.getenv("HUB_PASSWORD")
    hub_url = os.getenv("HUB_URL")

    if not hub_username or not hub_password or not hub_url:
        raise ValueError("HUB_USERNAME, HUB_PASSWORD, and HUB_URL environment variables must be set")

    async with Hub(hub_url) as hub:
        success, _, _ = await hub.signin(hub_username, hub_password)
        if not success:
            raise ConnectionError(f"Failed to authenticate with Hub.")            

        personas = await hub.list_personas(persona_module['name'])
    persona = personas[0]
    persona_url = persona['module_url']

    # Clone the repo
    repo_name = persona_url.split('/')[-1]
    repo_path = Path(f"{AGENT_DIR}/{repo_name}")
    
    # Remove existing repo if it exists
    if repo_path.exists():
        import shutil
        shutil.rmtree(repo_path)
        
    _ = Repo.clone_from(persona_url, to_path=str(repo_path))
    
    persona_file = repo_path / persona['module_entrypoint']
    if not persona_file.exists():
        logger.error(f"Persona file not found in repository {repo_name}")
        return None
            
    # Load based on file extension
    with persona_file.open('r') as f:
        if persona_file.suffix == '.json':
            persona_data = json.load(f)
        elif persona_file.suffix in ['.yml', '.yaml']:
            persona_data = yaml.safe_load(f)
        else:
            logger.error(f"Unsupported file type {persona_file.suffix} in {repo_name}")
            return None
        

    # input_schema = load_input_schema(repo_name)
    return persona_data
        
    
    
def read_gitignore(directory):
    gitignore_path = os.path.join(directory, '.gitignore')
    
    if not os.path.exists(gitignore_path):
        logger.info(f"No .gitignore file found in {directory}")
        return []
    
    with open(gitignore_path, 'r') as file:
        lines = file.readlines()

    ignored_files = [line.strip() for line in lines if line.strip() and not line.startswith('#')]
    return ignored_files

def zip_dir_with_gitignore(directory_path):
    ignored_files = read_gitignore(directory_path)
    output_zip_file = f"./{os.path.basename(directory_path)}.zip"

    # Convert patterns in .gitignore to absolute paths for comparison
    ignored_patterns = [os.path.join(directory_path, pattern) for pattern in ignored_files]

    with zipfile.ZipFile(output_zip_file, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for root, dirs, files in os.walk(directory_path):
            dirs = [d for d in dirs if not any(fnmatch.fnmatch(os.path.join(root, d), pattern) for pattern in ignored_patterns)]
            
            for file in files:
                file_path = os.path.join(root, file)

                if any(fnmatch.fnmatch(file_path, pattern) for pattern in ignored_patterns):
                    continue
                
                if file == output_zip_file.split('/')[1]:
                    continue

                zip_file.write(file_path, os.path.relpath(file_path, directory_path))

    logger.info(f"Zipped directory '{directory_path}' to '{output_zip_file}'")
    return output_zip_file



from naptha_sdk.client.node import NodeClient
from naptha_sdk.schemas import KBDeployment, KBRunInput
from typing import Any, Dict
import logging
logger = logging.getLogger(__name__)

class KnowledgeBase:
    def __init__(self, kb_deployment: KBDeployment):
        self.kb_deployment = kb_deployment
        self.kb_node = NodeClient(self.kb_deployment.node)
        self.table_name = kb_deployment.config.path
        self.schema = kb_deployment.config.schema
        if "id_column" in kb_deployment.config:
            self.id_column = kb_deployment.config.id_column
        else:
            self.id_column = "id"
        if self.table_name is None:
            self.table_name = kb_deployment.module["name"]

    @classmethod
    async def create(cls, module_run):
        instance = cls(module_run)
        await instance._initialize()
        return instance

    async def _initialize(self):
        try:
            tables = await self.kb_node.list_tables()
            if self.table_name not in tables:
                await self.kb_node.create_table(self.table_name, self.schema)
        except Exception as e:
            logger.error(f"Error initializing knowledge base: {str(e)}")
            raise
        
    async def call_kb_func(self, module_run_input: KBRunInput, *args, **kwargs):
        logger.info(f"Running knowledge base on knowledge base node {self.kb_node}")
        kb_run = await self.kb_node.run_module(module_type="kb", run_input=module_run_input.model_dict())
        return kb_run



import ast
import inspect
from naptha_sdk.module_manager import sort_modules, extract_dependencies
import os
from pathlib import Path
import sys
import typing
from typing import TypeVar
import yaml

def is_local_module(module):
    if not hasattr(module, '__file__'):
        return False  # Built-in modules don't have __file__

    module_path = os.path.abspath(module.__file__)
    current_dir = os.path.abspath(os.getcwd())

    # Check if the module is in the current directory or its subdirectories
    if module_path.startswith(current_dir):
        # Check if the module is in the .venv directory
        venv_dir = os.path.join(current_dir, '.venv')
        if module_path.startswith(venv_dir):
            return False  # It's in .venv, so it's an installed package
        return True  # It's in the project directory, but not in .venv
    
    return False  # It's outside the project directory

def scrape_init(file_path):
    def extract_value(value):
        if isinstance(value, ast.Constant):
            return value.value
        elif isinstance(value, ast.Name):
            return value.id
        elif isinstance(value, ast.Attribute):
            return f"{extract_value(value.value)}.{value.attr}"
        elif isinstance(value, ast.Call):
            if isinstance(value.func, ast.Attribute):
                return f"{extract_value(value.func.value)}.{value.func.attr}()"
            elif isinstance(value.func, ast.Name):
                return f"{value.func.id}()"
        else:
            return ast.unparse(value)

    with open(file_path, 'r') as file:
        tree = ast.parse(file.read(), filename=file_path)

    variables = []
    unique_variables = {}

    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    if isinstance(node.value, ast.Call) and isinstance(node.value.func, ast.Name):
                        data = {"type": "call", "target": target.id, "cls_name": node.value.func.id}
                        if node.value.keywords:
                            data['keywords'] = [kw.arg for kw in node.value.keywords]
                            data['values'] = [extract_value(kw.value) for kw in node.value.keywords]
                        unique_variables[data['target']] = data
                    elif isinstance(node.value, ast.Constant):
                        data = {"type": "constant", "target": target.id, "value": node.value.value}
                        unique_variables[data['target']] = data

    # Convert the dictionary values back to a list
    variables = list(unique_variables.values())

    return variables

def get_obj_dependencies(context_globals, fn_code, processed=None):
    if processed is None:
        processed = set()

    modules = []
    variables = []
    for name, obj in context_globals.items():
        if name in fn_code and obj not in processed:
            print("name", name, "obj", obj, type(obj), type(obj).__name__)
            if name.startswith("__"):
                continue

            processed.add(obj)  # Add the current object to the set of processed objects

            if "Union" in type(obj).__name__:
                print("Union", obj.__args__)
                args_str = ', '.join(arg.__name__ if hasattr(arg, '__name__') else str(arg) for arg in obj.__args__)
                variables.append({"type": "union", "target": name, "value": f"Union[{args_str}]"})
                for arg in obj.__args__:
                    if inspect.isclass(arg):
                        obj_info = {
                            'name': arg.__name__,
                            'module': arg.__module__,
                            'import_type': "selective",
                            'is_local': False
                        }
                        modules.append(obj_info)
            elif isinstance(obj, (int, str, float, bool)):
                variables.append({"type": "constant", "target": name, "value": f"{obj}"})
            elif inspect.ismodule(obj):
                obj_info = {
                    'name': name,
                    'module': obj.__name__,
                    'import_type': "standard",
                    'is_local': is_local_module(obj)
                }
                modules.append(obj_info)
            else:
                module = sys.modules.get(obj.__module__, None)
                if module and module not in processed:
                    is_local = is_local_module(module)
                    obj_info = {
                        'name': name,
                        'module': obj.__module__,
                        'import_type': "selective",
                        'is_local': is_local
                    }
                    if is_local:
                        if isinstance(obj, TypeVar):
                            obj_info['source'] = ""
                        else:
                            obj_info['source'] = inspect.getsource(obj)
                            new_modules, new_variables = get_obj_dependencies(module.__dict__, obj_info['source'], processed)
                            modules.extend(new_modules)
                            variables.extend(new_variables)
                    modules.append(obj_info)

    return modules, variables

def scrape_func_params(func):
    # Extract func parameter names, default values, and type annotations
    sig = inspect.signature(func)
    params = {}
    for param_name, param in sig.parameters.items():
        if param_name != "self":
            param_info = {
                'value': None if param.default is param.empty else param.default,
                'type': None
            }
            
            # Get type annotation if available
            if param.annotation is not param.empty:
                param_info['type'] = param.annotation
            
            params[param_name] = param_info
    
    return params

def scrape_func(func, variables):
    fn_code = inspect.getsource(func)
    fn_name = func.__name__
    fn_code = "\n".join(line for line in fn_code.splitlines() if not line.strip().startswith("@"))

    # check which variables are used in the function
    used_variables = []
    for variable in variables:
        if variable['target'] in fn_code:
            used_variables.append(variable)

    if inspect.isfunction(func):
        context_globals = func.__globals__
    elif inspect.isclass(func):
        module = sys.modules[func.__module__]
        context_globals = module.__dict__

    modules, new_variables = get_obj_dependencies(context_globals, fn_code)
    used_variables.extend(new_variables)

    # Deal with variables from the main file
    for used_variable in used_variables:
        if used_variable['type'] == 'constant':
            if isinstance(used_variable['value'], str) and used_variable['value'].endswith('.yaml'):
                cwd = Path.cwd()
                full_path = Path(f"src/{cwd.name}/{used_variable['value']}")
                if full_path.exists():
                    with open(full_path, 'r') as file:
                        yaml_data = yaml.safe_load(file)
                    line = f"{used_variable['target']} = {yaml_data}\n"
                    class_info = {
                        'name': used_variable['target'],
                        'module': None,
                        'import_type': "variable",
                        'is_local': False
                    }
                    class_info['source'] = line
            else:
                line = f"{used_variable['target']} = {used_variable['value']}\n"
                class_info = {
                    'name': used_variable['target'],
                    'module': None,
                    'import_type': "variable",
                    'is_local': False,
                    'source': line
                }
            modules.append(class_info)
        elif used_variable['type'] == 'union':
            line = f"{used_variable['target']} = {used_variable['value']}\n"
            class_info = {
                'name': used_variable['target'],
                'module': None,
                'import_type': "union",
                'is_local': False
            }
            class_info['source'] = line
            modules.append(class_info)
        elif used_variable['type'] == 'call':
            # If the variable's class is not in modules, add it
            var_class = context_globals.get(used_variable['cls_name'])
            if var_class and inspect.isclass(var_class):
                module = sys.modules[var_class.__module__]
                class_info = {
                    'name': used_variable['cls_name'],
                    'module': var_class.__module__,
                    'import_type': "variable",
                    'is_local': False
                }
                line = f"{used_variable['target']} = {used_variable['cls_name']}("
                if 'keywords' in used_variable:
                    for kw, value in zip(used_variable['keywords'], used_variable['values']):
                        if isinstance(value, str):
                            line += f'{kw}="{value}", '
                        else:
                            line += f'{kw}={value}, '
                line += ")\n"
                class_info['source'] = line
                if any(module['name'] == used_variable['cls_name'] for module in modules):
                    class_info['import_needed'] = False
                modules.append(class_info)

    modules = [module for module in modules if module['name'] != 'logger']
    local_modules = [module for module in modules if module['is_local']]
    module_dependencies = {mod['name']: extract_dependencies(mod, local_modules) for mod in local_modules}
    local_modules = sort_modules(local_modules, module_dependencies) # Sort local modules based on dependencies
    selective_import_modules = [module for module in modules if not module['is_local'] and module['import_type'] == 'selective']
    standard_import_modules = [module for module in modules if module['import_type'] == 'standard']
    variable_modules = [module for module in modules if module['import_type'] == 'variable']
    union_modules = [module for module in modules if module['import_type'] == 'union']

    return fn_code, fn_name, local_modules, selective_import_modules, standard_import_modules, variable_modules, union_modules
